
/*
 * Copyright 1992-2000 by Alan Hourihane, Wigan, England.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Alan Hourihane not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Alan Hourihane makes no representations
 * about the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 * ALAN HOURIHANE DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL ALAN HOURIHANE BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Alan Hourihane, alanh@fairlite.demon.co.uk
 */
/* $XFree86: xc/programs/Xserver/hw/xfree86/drivers/trident/trident_dac.c,v 1.62 2002/01/08 10:48:53 alanh Exp $ */

#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86_ansic.h"
#include "xf86Version.h"
#include "xf86PciInfo.h"
#include "xf86Pci.h"

#include "vgaHW.h"

#include "trident.h"
#include "trident_regs.h"

static biosMode bios1[] = {
    { 640, 480, 0x11 }
};

static biosMode bios4[] = {
    { 320, 200, 0xd },
    { 640, 200, 0xe },
    { 640, 350, 0x11 },
    { 640, 480, 0x12 },
    { 800, 600, 0x5b },
    { 1024, 768 , 0x5f },
    { 1280, 1024, 0x63 },
    { 1600, 1200, 0x65 }
};

static biosMode bios8[] = {	
    { 320, 200, 0x13 },
    { 640, 400, 0x5c },
    { 640, 480, 0x5d },
    { 720, 480, 0x60 },
    { 800, 600, 0x5e },
    { 1024, 768, 0x62 },
    { 1280, 1024, 0x64 },
    { 1600, 1200, 0x66 }
};

static biosMode bios15[] = {
    { 640, 400, 0x72 },
    { 640, 480, 0x74 },
    { 720, 480, 0x70 },
    { 800, 600, 0x76 },
    { 1024, 768, 0x78 },
    { 1280, 1024, 0x7a },
    { 1600, 1200, 0x7c }
};

static biosMode bios16[] = {
    { 640, 400, 0x73 },
    { 640, 480, 0x75 },
    { 720, 480, 0x71 },
    { 800, 600, 0x77 },
    { 1024, 768, 0x79 },
    { 1280, 1024, 0x7b },
    { 1600, 1200, 0x7d }
};

static biosMode bios24[] = {
    { 640, 400, 0x6b },
    { 640, 480, 0x6c },
    { 720, 480, 0x61 },
    { 800, 600, 0x6d },
    { 1024, 768, 0x6e }
};

static newModes newModeRegs [] = {
  { 320, 200, 0x13, 0x30 },
  { 640, 480, 0x13, 0x61 },
  { 800, 600, 0x13, 0x61 },
  { 1024, 768, 0x3b, 0x63 },
  { 1280, 1024, 0x7b, 0x64 }
};

/***************************************************************************
 *
 * TV parameters for VT1621
 *
 ***************************************************************************/
#define TV_MODE  12  /* Add 4 modes of PALM by ian */
#define TVX_MODE_SIZE 0X72
#define TVX_CRTC_NUM 0x10
#define TVX_REG_NUM 0x62
#define TVX_VT1621_PORT 0x40
#define SMBUS_BASE 0x5000
unsigned char TVX_VT1621_Table[TV_MODE][TVX_MODE_SIZE] = {
{
/* NTSC, 640x480, bpp=8,16 */
0x02, 0x54, 0xE0, 0xFA, 0x11, 0x5D, 0x11, 0x57, 0x5A, 0x56,
0xA0, 0x26, 0x0A, 0x55, 0x37, 0x86,

0x6A,0x0B,0x22,0x27,0x43,0x50,0x13,0x56,0xB0,0x07,0xF4,0x1B,0x5A,0xF4,0x00,0xB9,
0x00,0x00,0x0E,0x48,0x38,0x38,0x00,0x44,0x00,0x40,0x0C,0x02,0x01,0x80,0x00,0x00,
0x0F,0x09,0x99,0x78,0x04,0x5A,0x36,0x9C,0x5B,0x00,0x00,0xAD,0x2C,0x89,0xFD,0x00,
0x00,0x17,0x15,0x21,0x15,0x05,0x05,0x02,0x1B,0x1B,0x24,0xF8,0x07,0x00,0x00,0x0F,
0x0F,0x60,0x01,0x0A,0x00,0x05,0x04,0xFF,0x03,0x01,0x90,0x33,0x00,0x00,0x00,0x00,
0x00,0x04,0x3F,0x02,0x02,0xFD,0x06,0xF8,0x0B,0xF3,0x0F,0x70,0x05,0xF9,0x0B,0xF1,
0x11,0x6E
},
{
/* NTSC, 800x600, bpp=8,16 */
0X02, 0x79, 0XE0, 0x73, 0x02, 0x80, 0x01, 0x7A, 0x7E, 0xEC,
0xA0, 0x8A, 0x0E, 0xEB, 0x8B, 0x89,

0x8B,0x0B,0x6A,0x27,0x43,0x50,0x12,0x4E,0xC0,0x0A,0xF2,0x19,0x70,0xF2,0x00,0xA9,
0x00,0x00,0x0A,0x48,0x1C,0x28,0x03,0x49,0x00,0x40,0x36,0x02,0x03,0x80,0x00,0x00,
0x0D,0x06,0x04,0x78,0x00,0x5A,0x41,0x98,0x6B,0x00,0x00,0xAD,0x2C,0xA0,0x2E,0x10,
0x00,0x17,0x15,0x21,0x15,0x05,0x05,0x02,0x1B,0x1B,0x24,0xF8,0x07,0x00,0x00,0x0F,
0x0F,0x60,0x01,0x0A,0x00,0x05,0x04,0xFF,0x03,0x01,0xD6,0x80,0x00,0x00,0x00,0x00,
0x00,0x0C,0x3F,0x02,0x02,0xFD,0x06,0xF8,0x0B,0xF3,0x0F,0x70,0x05,0xF9,0x0B,0xF1,
0x11,0x6E
},
{
/* NTSC, 640x480, bpp=24 */
0X02, 0x54, 0XE0, 0xFA, 0x11, 0x5D, 0x11, 0x57, 0x5A, 0x56,
0xA0, 0x26, 0x0A, 0X55, 0x37, 0x46,

0x6A,0x0B,0x23,0x33,0x43,0x50,0x13,0x53,0xB0,0x07,0xD5,0x19,0x6B,0xD5,0x00,0xB9,
0x00,0x00,0x0E,0x48,0x38,0x38,0x03,0x44,0x00,0x40,0x0C,0x02,0x03,0x80,0x00,0x00,
0x0F,0x09,0x99,0x79,0x04,0x5A,0x36,0x96,0x5B,0x00,0x00,0xAD,0x2C,0x89,0xFD,0x00,
0x00,0x17,0x15,0x21,0x15,0x05,0x05,0x02,0x1B,0x1B,0x24,0xF8,0x07,0x00,0x00,0x0F,
0x0F,0x60,0x01,0x0A,0x00,0x05,0x04,0xFF,0x03,0x01,0xA0,0x33,0x1B,0x00,0x00,0x00,
0x00,0x08,0x3F,0x02,0x02,0xFD,0x06,0xF8,0x0B,0xF3,0x0F,0x70,0x05,0xF9,0x0B,0xF1,
0x11,0x6E
},
{
/* NTSC, 800x600, bpp=24 */
0X02, 0x79, 0XE0, 0x73, 0x02, 0x80, 0x01, 0x7B, 0x7E, 0xEC,
0xA0, 0x8A, 0x0E, 0xEB, 0x8B, 0x49,

0x8B,0x0B,0x6B,0x27,0x03,0x50,0x12,0x43,0xC0,0x0C,0x90,0x19,0x42,0x90,0x00,0x99,
0x00,0x00,0x0A,0x48,0x1C,0x28,0x00,0x4B,0x00,0x40,0x36,0x02,0x03,0x80,0x00,0x00,
0x0D,0x06,0x04,0x79,0x00,0x5A,0x41,0x98,0x6B,0x00,0x00,0xAA,0x2C,0xA0,0x2E,0x10,
0x00,0x17,0x15,0x21,0x15,0x05,0x05,0x02,0x1B,0x1B,0x24,0xF8,0x07,0x00,0x00,0x0F,
0x0F,0x60,0x01,0x0A,0x00,0x05,0x04,0xFF,0x03,0x01,0xC6,0x90,0x00,0x00,0x00,0x00,
0x00,0x08,0x3F,0x02,0x02,0xFD,0x06,0xF8,0x0B,0xF3,0x0F,0x70,0x05,0xF9,0x0B,0xF1,
0x11,0x6E
},
{
/* PAL, 640x480, bpp=8,16 */
0X82, 0x5D, 0XE0, 0x23, 0x02, 0x64, 0x11, 0x56, 0x5A, 0x6F,
0xA0, 0x0D, 0x0F, 0x6E, 0x24, 0xC1,

0x6b,0x0b,0x03,0x67,0x40,0x50,0x12,0x96,0xce,0x32,0xff,0x01,0x7e,0xf6,0x00,0xa8,
0x00,0x00,0x07,0x48,0x20,0x1c,0x44,0x60,0x44,0x4f,0x1b,0x02,0x03,0x80,0x00,0x00,
0x0c,0x0c,0xd7,0x84,0x04,0x68,0x3b,0x9c,0x57,0x63,0x17,0xac,0x25,0x80,0x29,0x10,
0x00,0x1a,0x22,0x2a,0x22,0x05,0x02,0x00,0x1c,0x3d,0x14,0xfe,0x03,0x54,0x01,0xfe,
0x7e,0x60,0x00,0x08,0x00,0x04,0x07,0x55,0x01,0x01,0xa0,0x33,0x00,0x00,0x00,0x00,
0x00,0x0c,0x4e,0xfe,0x03,0xfb,0x06,0xf8,0x0a,0xf5,0x0c,0x73,0x06,0xf8,0x0b,0xf2,
0x10,0x6f
},
{
/* PAL, 800x600, bpp=8,16 */
0X82, 0x5B, 0XE0, 0x91, 0x02, 0x73, 0x07, 0x6C, 0x70, 0xEC,
0xA0, 0xA8, 0x0B, 0xEB, 0xAD, 0xC7,

0x8b,0x0b,0x1a,0x47,0x40,0x50,0x12,0x56,0x00,0x37,0xf7,0x00,0x7d,0xe2,0x00,0xb8,
0x00,0x00,0x0e,0x48,0x38,0x38,0x44,0x62,0x44,0x4f,0x53,0x02,0x07,0x80,0x00,0x00,
0x0a,0x05,0xa2,0x83,0x08,0x68,0x46,0x99,0x68,0x63,0x17,0xac,0x25,0x80,0x6b,0x10,
0x00,0x1a,0x22,0x2a,0x22,0x05,0x02,0x00,0x1c,0x3d,0x14,0xfe,0x03,0x54,0x01,0xfe,
0x7e,0x60,0x00,0x08,0x00,0x04,0x07,0x55,0x01,0x01,0xe6,0x90,0x00,0x00,0x00,0x00,
0x00,0x0c,0x4d,0xfb,0x04,0xfb,0x07,0xf8,0x09,0xf6,0x0a,0x74,0x06,0xf8,0x0b,0xf2,
0x10,0x6f
},
{
/* PAL, 640x480, bpp=24 */
0X82, 0x5D, 0XE0, 0x23, 0x02, 0x64, 0x11, 0x56, 0x5A, 0x6F,
0xA0, 0x0D, 0x0F, 0x6E, 0x24, 0x81,

0x6b,0x0b,0x02,0x67,0x40,0x50,0x12,0x93,0xce,0x32,0xf0,0x01,0x88,0xe8,0x00,0xa8,
0x00,0x00,0x07,0x48,0x20,0x1c,0x44,0x60,0x44,0x4f,0x1b,0x02,0x03,0x80,0x00,0x00,
0x0c,0x05,0xe2,0x84,0x00,0x68,0x3b,0x9c,0x57,0x63,0x17,0xac,0x25,0x80,0x29,0x10,
0x00,0x1a,0x22,0x2a,0x22,0x05,0x02,0x00,0x1c,0x3d,0x14,0xfe,0x03,0x54,0x01,0xfe,
0x7e,0x60,0x00,0x08,0x00,0x04,0x07,0x55,0x01,0x01,0xa0,0x33,0x00,0x00,0x00,0x00,
0x00,0x0c,0x4e,0xfe,0x03,0xfb,0x06,0xf8,0x0a,0xf5,0x0c,0x73,0x06,0xf8,0x0b,0xf2,
0x10,0x6f
},
{
/* PAL, 800x600, bpp=24 */
0X82, 0x5B, 0XE0, 0x91, 0x02, 0x73, 0x07, 0x6C, 0x70, 0xEC,
0xA0, 0xA8, 0x0B, 0xEB, 0xAD, 0x87,

0x8b,0x0b,0x1a,0x67,0x40,0x50,0x12,0x53,0x00,0x37,0xee,0x00,0x83,0xeb,0x00,0xb8,
0x00,0x00,0x0e,0x48,0x38,0x38,0x44,0x62,0x44,0x4f,0x53,0x02,0x07,0x80,0x00,0x00,
0x0a,0x05,0x5e,0x83,0x08,0x68,0x46,0x99,0x68,0x63,0x17,0xac,0x25,0x80,0x6b,0x10,
0x00,0x1a,0x22,0x2a,0x22,0x05,0x02,0x00,0x1c,0x3d,0x14,0xfe,0x03,0x54,0x01,0xfe,
0x7e,0x60,0x00,0x08,0x00,0x04,0x07,0x55,0x01,0x01,0xa0,0x22,0x00,0x00,0x00,0x00,
0x00,0x0c,0x4d,0xfb,0x04,0xfb,0x07,0xf8,0x09,0xf6,0x0a,0x74,0x06,0xf8,0x0b,0xf2,
0x10,0x6f
},
{
/* PALM, 640x480, bpp=8,16 */
0X82, 0x54, 0XE0, 0xFA, 0x11, 0x5D, 0x11, 0x57, 0x5A, 0x56,
0xA0, 0x26, 0x0A, 0x55, 0x37, 0x86,

0x6A,0x0B,0x22,0x26,0x42,0x50,0x13,0x56,0xB0,0x07,0xAE,0x17,0x6F,0xBD,0x01,0x19,
0x00,0x00,0x0E,0x48,0x38,0x38,0x00,0x30,0xED,0x3F,0x0C,0x02,0x01,0x80,0x00,0x00,
0x0F,0x0B,0x99,0x7B,0x04,0x5A,0x36,0x9C,0x5B,0x00,0x00,0xAD,0x2C,0x89,0xFD,0x00,
0x00,0x17,0x15,0x21,0x15,0x05,0x05,0x02,0x1B,0x1B,0x24,0xF8,0x07,0x00,0x00,0x0F,
0x0F,0x60,0x01,0x0A,0x00,0x05,0x04,0xFF,0x03,0x01,0x90,0x33,0x00,0x00,0x00,0x00,
0x00,0x04,0x46,0xfe,0x03,0xfb,0x06,0xf8,0x0a,0xf5,0x0c,0x73,0x06,0xf8,0x0b,0xf2,
0x10,0x6f
},
{
/* PALM, 800x600, bpp=8,16 */
0X82, 0x79, 0XE0, 0x73, 0x02, 0x80, 0x01, 0x7A, 0x7E, 0xEC,
0xA0, 0x8A, 0x0E, 0xEB, 0x8B, 0x89,

0x8B,0x0B,0x6A,0x26,0x42,0x50,0x12,0x4E,0xC0,0x0A,0xC7,0x18,0x71,0xDE,0x00,0xFB,
0x00,0x00,0x0A,0x48,0x1C,0x28,0x00,0x30,0xED,0x3F,0x36,0x02,0x03,0x80,0x00,0x00,
0x0D,0x06,0x04,0x7A,0x00,0x5A,0x41,0x98,0x6B,0x00,0x00,0xAD,0x2C,0xA0,0x2E,0x10,
0x00,0x17,0x15,0x21,0x15,0x05,0x05,0x02,0x1B,0x1B,0x24,0xF8,0x07,0x00,0x00,0x0F,
0x0F,0x60,0x01,0x0A,0x00,0x05,0x04,0xFF,0x03,0x01,0xD6,0x80,0x00,0x00,0x00,0x85,
0x00,0x0C,0x45,0xfb,0x04,0xfb,0x07,0xf8,0x09,0xf6,0x0a,0x74,0x06,0xf8,0x0b,0xf2,
0x10,0x6f
},
{
/* PALM, 640x480, bpp=24 */
0X82, 0x54, 0XE0, 0xFA, 0x11, 0x5D, 0x11, 0x57, 0x5A, 0x56,
0xA0, 0x26, 0x0A, 0x55, 0x37, 0x46,

0x6A,0x0B,0x23,0x32,0x42,0x50,0x13,0x53,0xB0,0x07,0xAD,0x19,0x6D,0xB0,0x00,0x39,
0x00,0x00,0x0E,0x48,0x38,0x38,0x00,0x30,0xED,0x3F,0x0C,0x02,0x03,0x80,0x00,0x00,
0x0F,0x0B,0x99,0x7B,0x04,0x5A,0x36,0x96,0x5B,0x00,0x00,0xAD,0x2C,0x89,0xFD,0x00,
0x00,0x17,0x15,0x21,0x15,0x05,0x05,0x02,0x1B,0x1B,0x24,0xF8,0x07,0x00,0x00,0x0F,
0x0F,0x60,0x01,0x0A,0x00,0x05,0x04,0xFF,0x03,0x01,0xA0,0x33,0x1B,0x00,0x00,0x00,
0x00,0x08,0x46,0xfe,0x03,0xfb,0x06,0xf8,0x0a,0xf5,0x0c,0x73,0x06,0xf8,0x0b,0xf2,
0x10,0x6f
},
{
/* PALM, 800x600, bpp=24 */
0X82, 0x79, 0XE0, 0x73, 0x02, 0x80, 0x01, 0x7B, 0x7E, 0xEC,
0xA0, 0x8A, 0x0E, 0xEB, 0x8B, 0x49,

0x8B,0x0B,0x6B,0x26,0x02,0x50,0x12,0x43,0xC0,0x0C,0xA8,0x19,0x42,0x82,0x00,0xFB,
0x00,0x00,0x0A,0x48,0x1C,0x28,0x00,0x30,0xED,0x3F,0x36,0x02,0x03,0x80,0x00,0x00,
0x0D,0x06,0x04,0x7A,0x00,0x5A,0x41,0x98,0x6B,0x00,0x00,0xAA,0x2C,0xA0,0x2E,0x10,
0x00,0x17,0x15,0x21,0x15,0x05,0x05,0x02,0x1B,0x1B,0x24,0xF8,0x07,0x00,0x00,0x0F,
0x0F,0x60,0x01,0x0A,0x00,0x05,0x04,0xFF,0x03,0x01,0xC6,0x90,0x00,0x00,0x00,0x00,
0x00,0x08,0x45,0xfb,0x04,0xfb,0x07,0xf8,0x09,0xf6,0x0a,0x74,0x06,0xf8,0x0b,0xf2,
0x10,0x6f
}
};

/* VT1621 overscan parameters add by ian*/
#define VT1621OSCRCRNUM 13
#define VT1621OSCRSMNUM 16
typedef struct {
    unsigned char byCR[VT1621OSCRCRNUM];
    unsigned char bySM[VT1621OSCRSMNUM];
} tagVT1621OSCR;

/* if no set CR 0x2F to 0xBF, upper screen would be dark */
tagVT1621OSCR VT1621OScrRegIndex = {
    { 0xd0,0xd1,0xd2,0xd3,0xe0,0xe4,0xe5,0xe6,0xe7,0xf0,0xfe,0xff,0x2f },
    { 0x02,0x07,0x08,0x09,0x0a,0x0d,0x11,0x12,0x13,0x14,0x15,0x17,0x1c,0x23,0x24,0x4b }
};

tagVT1621OSCR VT1621OScrRegValue[TV_MODE] = {
    {
        { 0x54,0xe0,0xfa,0x11,0x5d,0x59,0x5b,0x0b,0x24,0xea,0xb9,0xcc,0xbf },/*640x480 8,16bpp ntsc*/
        { 0x03,0x52,0x80,0x00,0xF4,0xF4,0x07,0x0e,0x04,0x1c,0x28,0x12,0x03,0x7a,0x00,0x33 }
    },
    {
        { 0x79,0xe0,0x73,0x02,0x7d,0x70,0x75,0x74,0xa0,0x50,0xad,0x5f,0xbf },/*800x600 8,16bpp ntsc*/
        { 0x1a,0x84,0xb5,0xff,0xF2,0xF2,0x07,0x06,0x10,0x28,0x18,0x28,0x03,0x7a,0x0c,0x22 }
    },
    {
        { 0x54,0xe0,0xfa,0x11,0x5d,0x59,0x5b,0x0b,0x24,0xea,0xb9,0x8c,0xbf },/*640x480 24bpp ntsc*/
        { 0x02,0x51,0x7e,0x00,0xD5,0xD5,0x07,0x0e,0x04,0x1c,0x28,0x12,0x03,0x7a,0x00,0x33 }
    },
    {
        { 0x78,0xe0,0x12,0x08,0x7d,0x70,0x75,0x74,0xa0,0x59,0xad,0x1f,0xbf },/*800x600 24bpp ntsc*/
        { 0x1a,0x81,0xb5,0xf9,0x90,0x90,0x07,0x06,0x10,0x28,0x18,0x28,0x03,0x7a,0x0c,0x22 }
    },
    {
        { 0x6c,0xe0,0x23,0x02,0x73,0x64,0x70,0xf2,0xa5,0xe3,0xdc,0xcf,0xbf },/*640x480 8,16bpp pal*/
        { 0x12,0x7e,0x6f,0x1e,0xff,0xf6,0x00,0x0e,0x04,0x1c,0x28,0x76,0x03,0x84,0x08,0x33 }
    },
    {
        { 0x67,0xe0,0x65,0x12,0x71,0x6b,0x6d,0x6f,0xa0,0x5d,0x5f,0x57,0xbf },/*800x600 8,16bpp pal*/
        { 0x02,0x46,0x7e,0x3a,0xf7,0xe2,0x00,0x0e,0x04,0x1c,0x38,0x78,0x03,0x84,0x0c,0x33 }
    },
    {
        { 0x6c,0xe0,0x23,0x02,0x73,0x64,0x70,0xf2,0xa5,0xe3,0xdc,0x8f,0xbf },/*640x480 24bpp pal*/
        { 0x12,0x77,0x69,0x1e,0xf0,0xe8,0x00,0x0e,0x04,0x1c,0x28,0x76,0x03,0x84,0x0c,0x33 }
    },
    {
        { 0x67,0xe0,0x65,0x12,0x71,0x6b,0x6d,0x6f,0xa0,0x5d,0x5f,0x17,0xbf },/*800x600 24bpp pal*/
        { 0x02,0x47,0x7e,0x3a,0xee,0xeb,0x00,0x0e,0x04,0x1c,0x38,0x78,0x03,0x84,0x08,0x33 }
    }
};


/* TV Parameters for CH7005C */
#define TV_CH7005C_MODE_SIZE 45
#define TV_CH7005C_CRTC_NUM 0x10
#define TV_CH7005C_TVREG_NUM 29
#define TV_CH7005C_PORT 0xEA
unsigned char TV_CH7005C_Table[TV_MODE][TV_CH7005C_MODE_SIZE]={
{
/* NTSC 640x480 bpp=8,16 */
0x02, 0x80, 0x20, 0x02, 0x00, 0x5D, 0X80, 0X57, 0X80, 0X56,
0XBA, 0X10, 0X8C, 0X50, 0XF8, 0X7F,

0X6A, 0X7A, 0X00, 0X09, 0X80, 0X66, 0X00, 0X60,
0X2E, 0XFF, 0X01, 0X0B, 0X0C, 0X03, 0X40, 0X3F,
0X7E, 0X40, 0X02, 0X00, 0X08, 0X00, 0X00, 0X00,
0X00, 0X00, 0X0A, 0X02, 0X05
},
{
/* NTSC 800X600 bpp=8,16 */
0x02, 0x80, 0x20, 0x02, 0x00, 0x7D, 0X80, 0X6E, 0X1C, 0XBA,
0XF0, 0X70, 0X8C, 0XBA, 0X78, 0X53,

0X8C, 0X4A, 0X00, 0X09, 0X80, 0XAE, 0X01, 0X80,
0X2E, 0X02, 0X01, 0X0B, 0X7E, 0X7E, 0X7E, 0X7E,
0X7E, 0X40, 0X01, 0X0C, 0X00, 0X00, 0X00, 0X00,
0X00, 0X00, 0X0A, 0X00, 0X05
},
{
/* NTSC 640x480 bpp=24 */
0x02, 0x80, 0x20, 0x02, 0x00, 0x5D, 0X80, 0X57, 0X80, 0X56,
0XBA, 0X10, 0X8C, 0X50, 0XBD, 0X57,

0X6A, 0X7A, 0X00, 0X09, 0X80, 0X67, 0X00, 0X60,
0X2E, 0XFF, 0X01, 0X0B, 0X0C, 0X03, 0X40, 0X3F,
0X7E, 0X40, 0X02, 0X00, 0X07, 0X0C, 0X0D, 0X00,
0X00, 0X00, 0X0A, 0X02, 0X05
},
{
/* NTSC 800X600 bpp=24 */
0x02, 0x80, 0x20, 0x02, 0x00, 0x7D, 0X80, 0X6E, 0X1C, 0XBA,
0XF0, 0X70, 0X8C, 0XBA, 0XF8, 0X53,

0X8C, 0X4A, 0X00, 0X09, 0X80, 0XAF, 0X01, 0X80,
0X2E, 0X02, 0X01, 0X0B, 0X0C, 0X03, 0X40, 0X3F,
0X7E, 0X40, 0X01, 0X0C, 0X00, 0X00, 0X00, 0X00,
0X00, 0X00, 0X0A, 0X00, 0X05
},
{
/* PAL 640x480 bpp=8,16 */
0x82, 0x80, 0x20, 0x02, 0x00, 0x71, 0X74, 0X62, 0X84, 0X6F,
0XF0, 0X10, 0X09, 0XEB, 0X80, 0X5F,

0X81, 0X4A, 0X00, 0X09, 0X80, 0X84, 0X00, 0X70,
0X28, 0X02, 0X01, 0X0B, 0X0C, 0X03, 0X40, 0X3F,
0X7E, 0X40, 0X02, 0X00, 0X08, 0X00, 0X00, 0X00,
0X00, 0X00, 0X0A, 0X01, 0X05
},
{
/* PAL 800x600 bpp=8,16 */
0x82, 0x80, 0x20, 0x02, 0x00, 0x73, 0X76, 0X6A, 0X8C, 0XEC,
0XF0, 0X7E, 0X09, 0XEB, 0X8F, 0X8D,

0X83, 0X4A, 0X00, 0X09, 0X80, 0X7E, 0X00, 0X70,
0X3F, 0X02, 0X01, 0X0B, 0X0C, 0X03, 0X40, 0X3F,
0X7E, 0X40, 0X02, 0X00, 0X08, 0X00, 0X00, 0X00,
0X00, 0X00, 0X0A, 0X01, 0X05
},
{
/* PAL 640x480 bpp=24 */
0x82, 0x80, 0x20, 0x02, 0x00, 0x71, 0X74, 0X62, 0X84, 0X6F,
0XF0, 0X10, 0X09, 0XEB, 0X80, 0X1F,

0X81, 0X4A, 0X00, 0X09, 0X80, 0X84, 0X00, 0X70,
0X28, 0X02, 0X01, 0X0B, 0X0C, 0X03, 0X40, 0X3F,
0X7E, 0X40, 0X02, 0X00, 0X08, 0X00, 0X00, 0X00,
0X00, 0X00, 0X0A, 0X01, 0X05
},
{
/* PAL 800X600 bpp=24 */
0x82, 0x80, 0x20, 0x02, 0x00, 0x73, 0X76, 0X6A, 0X8C, 0XEC,
0XF0, 0X7E, 0X09, 0XEB, 0X5D, 0X48,

0X83, 0X4A, 0X00, 0X09, 0X80, 0X7E, 0X00, 0X70,
0X3F, 0X02, 0X01, 0X0B, 0X0C, 0X03, 0X40, 0X3F,
0X7E, 0X40, 0X02, 0X00, 0X08, 0X00, 0X00, 0X00,
0X00, 0X00, 0X0A, 0X01, 0X05
}
};

static void smbus_write(unsigned char bData, unsigned char bIndex, unsigned char devAdr);
static unsigned char smbus_read(unsigned char bIndex,unsigned char devAdr);


int
TridentFindMode(int xres, int yres, int depth)
{
    int xres_s;
    int i, size;
    biosMode *mode;

    switch (depth) {
    case 1:
	size = sizeof(bios1) / sizeof(biosMode);
	mode = bios1;
	break;
    case 4:
	size = sizeof(bios4) / sizeof(biosMode);
	mode = bios4;
	break;
    case 8:
	size = sizeof(bios8) / sizeof(biosMode);
	mode = bios8;
	break;
    case 15:
	size = sizeof(bios15) / sizeof(biosMode);
	mode = bios15;
	break;
    case 16:
	size = sizeof(bios16) / sizeof(biosMode);
	mode = bios16;
	break;
    case 24:
	size = sizeof(bios24) / sizeof(biosMode);
	mode = bios24;
	break;
    default:
	return 0;
    }

    for (i = 0; i < size; i++) {
	if (xres <= mode[i].x_res) {
	    xres_s = mode[i].x_res;
	    for (; i < size; i++) {
		if (mode[i].x_res != xres_s)
		    return mode[i-1].mode;
		if (yres <= mode[i].y_res)
		    return mode[i].mode;
	    }
	}
    }
    return mode[size - 1].mode;
}

static void
TridentFindNewMode(int xres, int yres, CARD8 *gr5a, CARD8 *gr5c)
{
    int xres_s;
    int i, size;
    
    size = sizeof(newModeRegs) / sizeof(newModes);

    for (i = 0; i < size; i++) {
	if (xres <= newModeRegs[i].x_res) {
	    xres_s = newModeRegs[i].x_res;
	    for (; i < size; i++) {
	        if (newModeRegs[i].x_res != xres_s 
		    || yres <= newModeRegs[i].y_res) {
		  *gr5a = newModeRegs[i].GR5a;
		  *gr5c = newModeRegs[i].GR5c;
		  return;
		}
	    }
	}
    }
    *gr5a = newModeRegs[size - 1].GR5a;
    *gr5c = newModeRegs[size - 1].GR5c;
    return;
}

Bool
TridentInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
{
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    TRIDENTRegPtr pReg = &pTrident->ModeReg;

    int vgaIOBase;
    int offset = 0;
    int clock = pTrident->currentClock;
    CARD8 protect;
    Bool fullSize = FALSE;
    Bool isShadow = FALSE;

    vgaHWPtr hwp = VGAHWPTR(pScrn);
    vgaRegPtr regp = &hwp->ModeReg;
    vgaRegPtr vgaReg = &hwp->ModeReg;
    vgaHWGetIOBase(VGAHWPTR(pScrn));
    vgaIOBase = VGAHWPTR(pScrn)->IOBase;

    /* Unprotect */
    OUTB(0x3C4, 0x11);
    protect = INB(0x3C5);
    OUTB(0x3C5, 0x92);

    OUTB(0x3C4, 0x0B); INB(0x3C5); /* Ensure we are in New Mode */

    pReg->tridentRegs3x4[PixelBusReg] = 0x00;
    pReg->tridentRegsDAC[0x00] = 0x00;
    pReg->tridentRegs3C4[NewMode2] = 0x20;
    OUTB(0x3CE, MiscExtFunc);
    pReg->tridentRegs3CE[MiscExtFunc] = INB(0x3CF) & 0xF0;
    pReg->tridentRegs3x4[GraphEngReg] = 0x00; 
    pReg->tridentRegs3x4[PreEndControl] = 0;
    pReg->tridentRegs3x4[PreEndFetch] = 0;

    pReg->tridentRegs3x4[CRTHiOrd] = (((mode->CrtcVBlankEnd-1) & 0x400)>>4) |
 				     (((mode->CrtcVTotal - 2) & 0x400) >> 3) |
 				     ((mode->CrtcVSyncStart & 0x400) >> 5) |
 				     (((mode->CrtcVDisplay - 1) & 0x400) >> 6)|
 				     0x08;

    pReg->tridentRegs3x4[HorizOverflow] = ((mode->CrtcHTotal & 0x800) >> 11) |
	    				  ((mode->CrtcHBlankStart & 0x800)>>7);

    if (pTrident->IsCyber) {
	Bool LCDActive;
#ifdef READOUT
	Bool ShadowModeActive;
#endif
	int i = pTrident->lcdMode;
#ifdef READOUT
	OUTB(0x3CE, CyberControl);
	ShadowModeActive = ((INB(0x3CF) & 0x81) == 0x81);
#endif
	OUTB(0x3CE, FPConfig);
	LCDActive = (INB(0x3CF) & 0x10);
	
	OUTB(0x3CE, CyberEnhance);
#if 0
	pReg->tridentRegs3CE[CyberEnhance] = INB(0x3CF);
#else
	pReg->tridentRegs3CE[CyberEnhance] = INB(0x3CF) & 0x8F;
	if (mode->CrtcVDisplay > 768)
	    pReg->tridentRegs3CE[CyberEnhance] |= 0x30;
	else
	if (mode->CrtcVDisplay > 600)
	    pReg->tridentRegs3CE[CyberEnhance] |= 0x20;
	else
	if (mode->CrtcVDisplay > 480)
	    pReg->tridentRegs3CE[CyberEnhance] |= 0x10;
#endif
	OUTB(0x3CE, CyberControl);
	pReg->tridentRegs3CE[CyberControl] = INB(0x3CF);

	OUTB(0x3CE,HorStretch);
	pReg->tridentRegs3CE[HorStretch] = INB(0x3CF);
	OUTB(0x3CE,VertStretch);
	pReg->tridentRegs3CE[VertStretch] = INB(0x3CF);

#ifdef READOUT
	if ((!((pReg->tridentRegs3CE[VertStretch] & 1) ||
	       (pReg->tridentRegs3CE[HorStretch] & 1)))
	    && (!LCDActive || ShadowModeActive)) 
	  {
	    unsigned char tmp;
	    
	    SHADOW_ENABLE(tmp);
	    OUTB(vgaIOBase + 4,0);
	    pReg->tridentRegs3x4[0x0] = INB(vgaIOBase + 5);
	    OUTB(vgaIOBase + 4,3);
	    pReg->tridentRegs3x4[0x3] = INB(vgaIOBase + 5);
	    OUTB(vgaIOBase + 4,4);
	    pReg->tridentRegs3x4[0x4] = INB(vgaIOBase + 5);
	    OUTB(vgaIOBase + 4,5);
  	    pReg->tridentRegs3x4[0x5] = INB(vgaIOBase + 5);
  	    OUTB(vgaIOBase + 4,0x6);
  	    pReg->tridentRegs3x4[0x6] = INB(vgaIOBase + 5);
  	    SHADOW_RESTORE(tmp);
 	} else
#endif
 	{
 	    if (i != 0xff) {
  		pReg->tridentRegs3x4[0x0] = LCD[i].shadow_0;
  		pReg->tridentRegs3x4[0x1] = regp->CRTC[1];
  		pReg->tridentRegs3x4[0x2] = regp->CRTC[2];
  		pReg->tridentRegs3x4[0x3] = LCD[i].shadow_3;
  		pReg->tridentRegs3x4[0x4] = LCD[i].shadow_4;
  		pReg->tridentRegs3x4[0x5] = LCD[i].shadow_5;
  		pReg->tridentRegs3x4[0x6] = LCD[i].shadow_6;
 		xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,
 			       "Overriding Horizontal timings.\n");
  	    }
  	}
 
 	if (i != 0xff) {
 	    pReg->tridentRegs3x4[0x7] = LCD[i].shadow_7;
 	    pReg->tridentRegs3x4[0x10] = LCD[i].shadow_10;
 	    pReg->tridentRegs3x4[0x11] = LCD[i].shadow_11;
 	    pReg->tridentRegs3x4[0x12] = regp->CRTC[0x12];
 	    pReg->tridentRegs3x4[0x15] = regp->CRTC[0x15];
 	    pReg->tridentRegs3x4[0x16] = LCD[i].shadow_16;
 	    if (LCDActive) {
 		pReg->tridentRegs3x4[CRTHiOrd] = LCD[i].shadow_HiOrd;
	    }
	    
	    fullSize = (pScrn->currentMode->HDisplay == LCD[i].display_x)
	        && (pScrn->currentMode->VDisplay == LCD[i].display_y);
 	}
 	
  	/* copy over common bits from normal VGA */
  	
  	pReg->tridentRegs3x4[0x7] &= ~0x4A;
	pReg->tridentRegs3x4[0x7] |= (vgaReg->CRTC[0x7] & 0x4A);

	if (LCDActive && fullSize) {	
	    regp->CRTC[0] = pReg->tridentRegs3x4[0];
	    regp->CRTC[3] = pReg->tridentRegs3x4[3];
	    regp->CRTC[4] = pReg->tridentRegs3x4[4];
	    regp->CRTC[5] = pReg->tridentRegs3x4[5];
	    regp->CRTC[6] = pReg->tridentRegs3x4[6];
	    regp->CRTC[7] = pReg->tridentRegs3x4[7];
	    regp->CRTC[0x10] = pReg->tridentRegs3x4[0x10];
	    regp->CRTC[0x11] = pReg->tridentRegs3x4[0x11];
	    regp->CRTC[0x16] = pReg->tridentRegs3x4[0x16];
	}
	if (LCDActive && !fullSize) {
	    /*
	     * Set negative h/vsync polarity to center display nicely
	     * Seems to work on several systems.
	     */
	    regp->MiscOutReg |= 0xC0;
	  /* 
	   * If the LCD is active and we don't fill the entire screen
	   * and the previous mode was stretched we may need help from
	   * the BIOS to set all registers for the unstreched mode.
	   */
	    pTrident->doInit =  ((pReg->tridentRegs3CE[HorStretch] & 1)
				|| (pReg->tridentRegs3CE[VertStretch] & 1));
	    pReg->tridentRegs3CE[CyberControl] |= 0x81;
	    xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"Shadow on\n");
	    isShadow = TRUE;
	} else {
	    pReg->tridentRegs3CE[CyberControl] &= 0x7E;
	    xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"Shadow off\n");
	}


	if (pTrident->CyberShadow) {
	    pReg->tridentRegs3CE[CyberControl] &= 0x7E;
	    isShadow = FALSE;
	    xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"Forcing Shadow off\n");
	}

 	xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"H-timing shadow registers:"
 		       " 0x%2.2x           0x%2.2x 0x%2.2x 0x%2.2x\n",
 		       pReg->tridentRegs3x4[0], pReg->tridentRegs3x4[3],
 		       pReg->tridentRegs3x4[4], pReg->tridentRegs3x4[5]);
 	xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"H-timing registers:       "
 		       " 0x%2.2x 0x%2.2x 0x%2.2x 0x%2.2x 0x%2.2x 0x%2.2x\n",
 		       regp->CRTC[0], regp->CRTC[1], regp->CRTC[2],
		       regp->CRTC[3], regp->CRTC[4], regp->CRTC[5]);
 	xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"V-timing shadow registers: "
 		       "0x%2.2x 0x%2.2x 0x%2.2x 0x%2.2x"
		       "           0x%2.2x (0x%2.2x)\n",
		       pReg->tridentRegs3x4[6], pReg->tridentRegs3x4[7],
 		       pReg->tridentRegs3x4[0x10],pReg->tridentRegs3x4[0x11],
 		       pReg->tridentRegs3x4[0x16],
 		       pReg->tridentRegs3x4[CRTHiOrd]);
 	xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"V-timing registers:        "
 		       "0x%2.2x 0x%2.2x 0x%2.2x 0x%2.2x "
		       "0x%2.2x 0x%2.2x 0x%2.2x\n",
 		       regp->CRTC[6], regp->CRTC[7], regp->CRTC[0x10],
		       regp->CRTC[0x11],regp->CRTC[0x12],
 		       regp->CRTC[0x14],regp->CRTC[0x16]);
 	
	
	/* disable stretching, enable centering */
	pReg->tridentRegs3CE[VertStretch] &= 0xFC;
	pReg->tridentRegs3CE[VertStretch] |= 0x80;
	pReg->tridentRegs3CE[HorStretch] &= 0xFC;
	pReg->tridentRegs3CE[HorStretch] |= 0x80;
#if 1
	{
  	    int mul = pScrn->bitsPerPixel >> 3; 
	    int val;
	    
	    if (!mul) mul = 1;
	    
	    /* this is what my BIOS does */ 
	    val = (pScrn->currentMode->HDisplay * mul / 8) + 16;

	    pReg->tridentRegs3x4[PreEndControl] = ((val >> 8) < 2 ? 2 :0)
	      | ((val >> 8) & 0x01);
	    pReg->tridentRegs3x4[PreEndFetch] = val & 0xff;
	}
#else
	OUTB(vgaIOBase + 4,PreEndControl);
	pReg->tridentRegs3x4[PreEndControl] = INB(vgaIOBase + 5);
	OUTB(vgaIOBase + 4,PreEndFetch);
	pReg->tridentRegs3x4[PreEndFetch] = INB(vgaIOBase + 5);
#endif
	/* set mode */
	if (pTrident->Chipset < CYBERBLADEXPm8) {
	  pReg->tridentRegs3CE[BiosMode] = TridentFindMode(
					   pScrn->currentMode->HDisplay,
					   pScrn->currentMode->VDisplay,
					   pScrn->depth);
	  xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, 1, 
			 "Setting BIOS Mode: %x\n",
			 pReg->tridentRegs3CE[BiosMode]);
	} else {
	  TridentFindNewMode(pScrn->currentMode->HDisplay,
			     pScrn->currentMode->VDisplay,
			     &pReg->tridentRegs3CE[BiosNewMode1],
			     &pReg->tridentRegs3CE[BiosNewMode2]);
	  xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, 1, 
			 "Setting BIOS Mode Regs: %x %x\n",
			 pReg->tridentRegs3CE[BiosNewMode1],
			 pReg->tridentRegs3CE[BiosNewMode2]);
	};
	
	/* no stretch */
	pReg->tridentRegs3CE[BiosReg] = 0;

	if (pTrident->CyberStretch) {
	    pReg->tridentRegs3CE[VertStretch] |= 0x01;
	    pReg->tridentRegs3CE[HorStretch] |= 0x01;
	    xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"Enabling StretchMode\n");
	}
    }

    /* Enable Chipset specific options */
    switch (pTrident->Chipset) {
	case CYBERBLADEXPAI1:
	case CYBERBLADEXPm8:
	case CYBERBLADEXPm16:
	case CYBERBLADEI7:
	case CYBERBLADEI7D:
	case CYBERBLADEI1:
	case CYBERBLADEI1D:
	case CYBERBLADEAI1:
	case CYBERBLADEAI1D:
	case CYBERBLADEE4:
	case BLADE3D:
	    OUTB(vgaIOBase + 4, RAMDACTiming);
	    pReg->tridentRegs3x4[RAMDACTiming] = INB(vgaIOBase + 5) | 0x0F;
	    /* Fall Through */
	case CYBER9520:
	case CYBER9525DVD:
	case CYBER9397DVD:
	case CYBER9397:
	case IMAGE975:
	case IMAGE985:
	case CYBER9388:
	    if (pScrn->bitsPerPixel >= 8)
    	    	pReg->tridentRegs3CE[MiscExtFunc] |= 0x10;
	    else
    	    	pReg->tridentRegs3CE[MiscExtFunc] &= ~0x10;
	    if (!pReg->tridentRegs3x4[PreEndControl])
	    	pReg->tridentRegs3x4[PreEndControl] = 0x01;
	    if (!pReg->tridentRegs3x4[PreEndFetch])
	    	pReg->tridentRegs3x4[PreEndFetch] = 0xFF;
	    /* Fall Through */
	case PROVIDIA9685:
	case CYBER9385:
	    pReg->tridentRegs3x4[Enhancement0] = 0x40;
	    /* Fall Through */
	case PROVIDIA9682:
	case CYBER9382:
	    if (pTrident->UsePCIRetry)
	    	pReg->tridentRegs3x4[PCIRetry] = 0xDF;
	    else
	    	pReg->tridentRegs3x4[PCIRetry] = 0x1F;
	    /* Fall Through */
	case TGUI9660:
	case TGUI9680:
	    if (pTrident->MUX && pScrn->bitsPerPixel == 8) {
	    	pReg->tridentRegs3x4[PixelBusReg] |= 0x01; /* 16bit bus */
	    	pReg->tridentRegs3C4[NewMode2] |= 0x02; /* half clock */
    		pReg->tridentRegsDAC[0x00] |= 0x20;	/* mux mode */
	    }	
    }


    /* Defaults for all trident chipsets follows */
    switch (pScrn->bitsPerPixel) {
	case 1:
	case 4:
    	    offset = pScrn->displayWidth >> 4;
	    break;
	case 8:
	    pReg->tridentRegs3CE[MiscExtFunc] |= 0x02;
    	    offset = pScrn->displayWidth >> 3;
	    break;
	case 16:
	    pReg->tridentRegs3CE[MiscExtFunc] |= 0x02;
    	    offset = pScrn->displayWidth >> 2;
	    if (pScrn->depth == 15)
    	    	pReg->tridentRegsDAC[0x00] = 0x10;
	    else
	    	pReg->tridentRegsDAC[0x00] = 0x30;
    	    pReg->tridentRegs3x4[PixelBusReg] = 0x04;
	    /* Reload with any chipset specific stuff here */
	    if (pTrident->Chipset >= TGUI9660) 
		pReg->tridentRegs3x4[PixelBusReg] |= 0x01;
	    if (pTrident->Chipset == TGUI9440AGi) {
    	        pReg->tridentRegs3CE[MiscExtFunc] |= 0x08;/*Clock Division / 2*/
	        clock *= 2;	/* Double the clock */
	    }
	    break;
	case 24:
	    pReg->tridentRegs3CE[MiscExtFunc] |= 0x02;
    	    offset = (pScrn->displayWidth * 3) >> 3;
    	    pReg->tridentRegs3x4[PixelBusReg] = 0x29;
	    pReg->tridentRegsDAC[0x00] = 0xD0;
	    if (pTrident->Chipset == CYBERBLADEE4) {
    		OUTB(vgaIOBase+ 4, New32);
		pReg->tridentRegs3x4[New32] = INB(vgaIOBase + 5) & 0x7F;
	    }
	    break;
	case 32:
	    pReg->tridentRegs3CE[MiscExtFunc] |= 0x02;
	    if (pTrident->Chipset != CYBERBLADEE4) {
	        /* Clock Division by 2*/
	        pReg->tridentRegs3CE[MiscExtFunc] |= 0x08; 
		clock *= 2;	/* Double the clock */
	    }
    	    offset = pScrn->displayWidth >> 1;
    	    pReg->tridentRegs3x4[PixelBusReg] = 0x09;
	    pReg->tridentRegsDAC[0x00] = 0xD0;
	    if (pTrident->Chipset == CYBERBLADEE4) {
    		OUTB(vgaIOBase+ 4, New32);
		pReg->tridentRegs3x4[New32] = INB(vgaIOBase + 5) | 0x80;
		/* With new mode 32bpp we set the packed flag */
    	    	pReg->tridentRegs3x4[PixelBusReg] |= 0x20;
	    }
	    break;
    }
    pReg->tridentRegs3x4[Offset] = offset & 0xFF;

    {
	CARD8 a, b;
	TGUISetClock(pScrn, clock, &a, &b);
	pReg->tridentRegsClock[0x00] = (regp->MiscOutReg & 0xF3) | 0x08;
	pReg->tridentRegsClock[0x01] = a;
	pReg->tridentRegsClock[0x02] = b;
	if (pTrident->MCLK > 0) {
	    TGUISetMCLK(pScrn, pTrident->MCLK, &a, &b);
	    pReg->tridentRegsClock[0x03] = a;
	    pReg->tridentRegsClock[0x04] = b;
	}
    }

    pReg->tridentRegs3C4[NewMode1] = 0xC0;
    pReg->tridentRegs3C4[Protection] = 0x92;

    pReg->tridentRegs3x4[LinearAddReg] = 0;
    if (pTrident->Linear) {
	/* This is used for VLB, when we support it again in 4.0 */
	if (pTrident->Chipset < CYBER9385)
    	    pReg->tridentRegs3x4[LinearAddReg] |=
					((pTrident->FbAddress >> 24) << 6)|
					((pTrident->FbAddress >> 20) & 0x0F);
	/* Turn on linear mapping */
    	pReg->tridentRegs3x4[LinearAddReg] |= 0x20; 
    } else {
	pReg->tridentRegs3CE[MiscExtFunc] |= 0x04;
    }

    pReg->tridentRegs3x4[CRTCModuleTest] =
				(mode->Flags & V_INTERLACE ? 0x84 : 0x80);

    OUTB(vgaIOBase+ 4, InterfaceSel);
    pReg->tridentRegs3x4[InterfaceSel] = INB(vgaIOBase + 5) | 0x40;

    OUTB(vgaIOBase+ 4, Performance);
    pReg->tridentRegs3x4[Performance] = INB(vgaIOBase + 5);

    if (pTrident->Chipset < CYBERBLADEXPm8)
	pReg->tridentRegs3x4[Performance] |= 0x10;

    OUTB(vgaIOBase+ 4, DRAMControl);
    pReg->tridentRegs3x4[DRAMControl] = INB(vgaIOBase + 5) | 0x10;

    if (pTrident->IsCyber && !pTrident->MMIOonly)
	pReg->tridentRegs3x4[DRAMControl] |= 0x20;

    OUTB(vgaIOBase+ 4, AddColReg);
    pReg->tridentRegs3x4[AddColReg] = INB(vgaIOBase + 5) & 0xEF;
    pReg->tridentRegs3x4[AddColReg] |= (offset & 0x100) >> 4;

    if (pTrident->Chipset >= TGUI9660) {
    	pReg->tridentRegs3x4[AddColReg] &= 0xDF;
    	pReg->tridentRegs3x4[AddColReg] |= (offset & 0x200) >> 4;
    }
   
    if (IsPciCard && UseMMIO) {
    	if (!pTrident->NoAccel)
	    pReg->tridentRegs3x4[GraphEngReg] |= 0x80;
    } else {
    	if (!pTrident->NoAccel)
	    pReg->tridentRegs3x4[GraphEngReg] |= 0x82; 
    }

    OUTB(0x3CE, MiscIntContReg);
    pReg->tridentRegs3CE[MiscIntContReg] = INB(0x3CF) | 0x04;

    /* Fix hashing problem in > 8bpp on 9320 chipset */
    if (pTrident->Chipset == CYBER9320 && pScrn->bitsPerPixel > 8) 
    	pReg->tridentRegs3CE[MiscIntContReg] &= ~0x80;

    OUTB(vgaIOBase+ 4, PCIReg);
    if (IsPciCard && UseMMIO)
    	pReg->tridentRegs3x4[PCIReg] = INB(vgaIOBase + 5) & 0xF9;
    else
    	pReg->tridentRegs3x4[PCIReg] = INB(vgaIOBase + 5) & 0xF8; 

    /* Enable PCI Bursting on capable chips */
    if (pTrident->Chipset >= TGUI9660) {
	if(pTrident->UsePCIBurst) {
	    pReg->tridentRegs3x4[PCIReg] |= 0x06;
	} else {
	    pReg->tridentRegs3x4[PCIReg] &= 0xF9;
	}
    }

    /* Video */
    OUTB(0x3C4,0x20);
    pReg->tridentRegs3C4[SSetup] = INB(0x3C5) | 0x4;
    pReg->tridentRegs3C4[SKey] = 0x00;
    pReg->tridentRegs3C4[SPKey] = 0xC0;

     /* restore */
    OUTB(0x3C4, 0x11);
    OUTB(0x3C5, protect);
   
    return(TRUE);
}

void
TridentRestore(ScrnInfoPtr pScrn, TRIDENTRegPtr tridentReg)
{
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    CARD8 temp;
    int vgaIOBase;
    vgaHWGetIOBase(VGAHWPTR(pScrn));
    vgaIOBase = VGAHWPTR(pScrn)->IOBase;

    if (pTrident->Chipset > PROVIDIA9685) {
    	OUTB(0x3C4, Protection);
    	OUTB(0x3C5, 0x92);
    }
#if 0
    if (pTrident->doInit && pTrident->Int10) {
        OUTW_3CE(BiosReg);	
    }
#endif
    /* Goto New Mode */
    OUTB(0x3C4, 0x0B);
    temp = INB(0x3C5);

    /* Unprotect registers */
    OUTW(0x3C4, ((0xC0 ^ 0x02) << 8) | NewMode1);
    
    temp = INB(0x3C8);
    temp = INB(0x3C6);
    temp = INB(0x3C6);
    temp = INB(0x3C6);
    temp = INB(0x3C6);
    OUTB(0x3C6, tridentReg->tridentRegsDAC[0x00]);
    temp = INB(0x3C8);

    OUTW_3x4(CRTCModuleTest);
    OUTW_3x4(LinearAddReg);
    OUTW_3C4(NewMode2);
    OUTW_3x4(CursorControl);
    OUTW_3x4(CRTHiOrd);
    OUTW_3x4(HorizOverflow);
    OUTW_3x4(AddColReg);
    OUTW_3x4(GraphEngReg);
    OUTW_3x4(Performance);
    OUTW_3x4(InterfaceSel);
    OUTW_3x4(DRAMControl);
    OUTW_3x4(PixelBusReg);
    OUTW_3x4(PCIReg);
    OUTW_3x4(PCIRetry);
    OUTW_3CE(MiscIntContReg);
    OUTW_3CE(MiscExtFunc);
    OUTW_3x4(Offset);
    OUTW_3C4(SSetup);
    OUTW_3C4(SKey);
    OUTW_3C4(SPKey);
    OUTW_3x4(PreEndControl);
    OUTW_3x4(PreEndFetch);
    if (pTrident->Chipset >= CYBER9385)    OUTW_3x4(Enhancement0);
    if (pTrident->Chipset >= BLADE3D)      OUTW_3x4(RAMDACTiming);
    if (pTrident->Chipset == CYBERBLADEE4) OUTW_3x4(New32);
    if (pTrident->IsCyber) {
	CARD8 tmp;

	OUTW_3CE(VertStretch);
	OUTW_3CE(HorStretch);
	if (pTrident->Chipset < CYBERBLADEXPm8) {
	    OUTW_3CE(BiosMode);
	} else {
	    OUTW_3CE(BiosNewMode1);
	    OUTW_3CE(BiosNewMode2);
	};
	OUTW_3CE(BiosReg);
    	OUTW_3CE(CyberControl);
    	OUTW_3CE(CyberEnhance);
	SHADOW_ENABLE(tmp);
	OUTW_3x4(0x0);
	if (pTrident->shadowNew) {
	    OUTW_3x4(0x1);
	    OUTW_3x4(0x2);	    
	}
	OUTW_3x4(0x3);
	OUTW_3x4(0x4);
	OUTW_3x4(0x5);
	OUTW_3x4(0x6);
	OUTW_3x4(0x7);	
	OUTW_3x4(0x10);
	OUTW_3x4(0x11);
	if (pTrident->shadowNew) {
	    OUTW_3x4(0x12);
	    OUTW_3x4(0x15);
	}
	OUTW_3x4(0x16);
	SHADOW_RESTORE(tmp);
    }
 
    if (Is3Dchip) {
#ifdef READOUT
	if (!pTrident->DontSetClock)
#endif
	{
	    OUTW(0x3C4, (tridentReg->tridentRegsClock[0x01])<<8 | ClockLow);
	    OUTW(0x3C4, (tridentReg->tridentRegsClock[0x02])<<8 | ClockHigh);
	}
	if (pTrident->MCLK > 0) {
	    OUTW(0x3C4,(tridentReg->tridentRegsClock[0x03])<<8 | MCLKLow);
	    OUTW(0x3C4,(tridentReg->tridentRegsClock[0x04])<<8 | MCLKHigh);
	}
    } else {
#ifdef READOUT
	if (!pTrident->DontSetClock)
#endif
	{
	    OUTB(0x43C8, tridentReg->tridentRegsClock[0x01]);
	    OUTB(0x43C9, tridentReg->tridentRegsClock[0x02]);
	}
	if (pTrident->MCLK > 0) {
	    OUTB(0x43C6, tridentReg->tridentRegsClock[0x03]);
	    OUTB(0x43C7, tridentReg->tridentRegsClock[0x04]);
	}
    }
#ifdef READOUT
    if (!pTrident->DontSetClock)
#endif
    {
	OUTB(0x3C2, tridentReg->tridentRegsClock[0x00]);
    }
    
    if (pTrident->Chipset > PROVIDIA9685) {
    	OUTB(0x3C4, Protection);
    	OUTB(0x3C5, tridentReg->tridentRegs3C4[Protection]);
    }

    OUTW(0x3C4, ((tridentReg->tridentRegs3C4[NewMode1] ^ 0x02) << 8)| NewMode1);
}

void
TridentSave(ScrnInfoPtr pScrn, TRIDENTRegPtr tridentReg)
{
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    CARD8 temp;
    int vgaIOBase;
    vgaHWGetIOBase(VGAHWPTR(pScrn));
    vgaIOBase = VGAHWPTR(pScrn)->IOBase;

    /* Goto New Mode */
    OUTB(0x3C4, 0x0B);
    temp = INB(0x3C5);

    INB_3C4(NewMode1);
    if (pTrident->Chipset > PROVIDIA9685)
    	INB_3C4(Protection);

    /* Unprotect registers */
    OUTW(0x3C4, ((0xC0 ^ 0x02) << 8) | NewMode1);
    OUTW(0x3C4, (0x92 << 8) | Protection);

    INB_3x4(Offset);
    INB_3x4(LinearAddReg);
    INB_3x4(CRTCModuleTest);
    INB_3x4(CRTHiOrd);
    INB_3x4(HorizOverflow);
    INB_3x4(Performance);
    INB_3x4(InterfaceSel);
    INB_3x4(DRAMControl);
    INB_3x4(AddColReg);
    INB_3x4(PixelBusReg);
    INB_3x4(GraphEngReg);
    INB_3x4(PCIReg);
    INB_3x4(PCIRetry);
    INB_3C4(SSetup);
    INB_3C4(SKey);
    INB_3C4(SPKey);
    INB_3x4(PreEndControl);
    INB_3x4(PreEndFetch);
    if (pTrident->Chipset >= CYBER9385)    INB_3x4(Enhancement0);
    if (pTrident->Chipset >= BLADE3D)      INB_3x4(RAMDACTiming);
    if (pTrident->Chipset == CYBERBLADEE4) INB_3x4(New32);
    if (pTrident->IsCyber) {
	CARD8 tmp;
	INB_3CE(VertStretch);
	INB_3CE(HorStretch);
	INB_3CE(BiosMode);
	INB_3CE(BiosNewMode1);
	INB_3CE(BiosNewMode2);
	INB_3CE(BiosReg);	
    	INB_3CE(CyberControl);
    	INB_3CE(CyberEnhance);
	SHADOW_ENABLE(tmp);
	INB_3x4(0x0);
	if (pTrident->shadowNew) {
	    INB_3x4(0x1);
	    INB_3x4(0x2);	    
	}
	INB_3x4(0x3);
	INB_3x4(0x4);
	INB_3x4(0x5);
	INB_3x4(0x6);
	INB_3x4(0x7);
	INB_3x4(0x10);
	INB_3x4(0x11);
	if (pTrident->shadowNew) {
	    INB_3x4(0x12);
	    INB_3x4(0x15);
	}
	INB_3x4(0x16);
	SHADOW_RESTORE(tmp);
    }

    /* save cursor registers */
    INB_3x4(CursorControl);

    INB_3CE(MiscExtFunc);
    INB_3CE(MiscIntContReg);

    temp = INB(0x3C8);
    temp = INB(0x3C6);
    temp = INB(0x3C6);
    temp = INB(0x3C6);
    temp = INB(0x3C6);
    tridentReg->tridentRegsDAC[0x00] = INB(0x3C6);
    temp = INB(0x3C8);

    tridentReg->tridentRegsClock[0x00] = INB(0x3CC);
    if (Is3Dchip) {
	OUTB(0x3C4, ClockLow);
	tridentReg->tridentRegsClock[0x01] = INB(0x3C5);
	OUTB(0x3C4, ClockHigh);
	tridentReg->tridentRegsClock[0x02] = INB(0x3C5);
	if (pTrident->MCLK > 0) {
	    OUTB(0x3C4, MCLKLow);
	    tridentReg->tridentRegsClock[0x03] = INB(0x3C5);
	    OUTB(0x3C4, MCLKHigh);
	    tridentReg->tridentRegsClock[0x04] = INB(0x3C5);
	}
    } else {
	tridentReg->tridentRegsClock[0x01] = INB(0x43C8);
	tridentReg->tridentRegsClock[0x02] = INB(0x43C9);
	if (pTrident->MCLK > 0) {
	    tridentReg->tridentRegsClock[0x03] = INB(0x43C6);
	    tridentReg->tridentRegsClock[0x04] = INB(0x43C7);
	}
    }

    INB_3C4(NewMode2);

    /* Protect registers */
    OUTW_3C4(NewMode1);
}

static void 
TridentShowCursor(ScrnInfoPtr pScrn) 
{
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    int vgaIOBase;
    vgaHWGetIOBase(VGAHWPTR(pScrn));
    vgaIOBase = VGAHWPTR(pScrn)->IOBase;

    /* 64x64 */
    OUTW(vgaIOBase + 4, 0xC150);
}

static void 
TridentHideCursor(ScrnInfoPtr pScrn) {
    int vgaIOBase;
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    vgaHWGetIOBase(VGAHWPTR(pScrn));
    vgaIOBase = VGAHWPTR(pScrn)->IOBase;

    OUTW(vgaIOBase + 4, 0x4150);
}

static void 
TridentSetCursorPosition(ScrnInfoPtr pScrn, int x, int y)
{
    int vgaIOBase;
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    vgaHWGetIOBase(VGAHWPTR(pScrn));
    vgaIOBase = VGAHWPTR(pScrn)->IOBase;
    
    if (x < 0) {
    	OUTW(vgaIOBase + 4, (-x)<<8 | 0x46);
	x = 0;
    } else
    	OUTW(vgaIOBase + 4, 0x0046);

    if (y < 0) {
    	OUTW(vgaIOBase + 4, (-y)<<8 | 0x47);
	y = 0;
    } else
    	OUTW(vgaIOBase + 4, 0x0047);

    OUTW(vgaIOBase + 4, (x&0xFF)<<8 | 0x40);
    OUTW(vgaIOBase + 4, (x&0x0F00)  | 0x41);
    OUTW(vgaIOBase + 4, (y&0xFF)<<8 | 0x42);
    OUTW(vgaIOBase + 4, (y&0x0F00)  | 0x43);
}

static void
TridentSetCursorColors(ScrnInfoPtr pScrn, int bg, int fg)
{
    int vgaIOBase;
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    vgaHWGetIOBase(VGAHWPTR(pScrn));
    vgaIOBase = VGAHWPTR(pScrn)->IOBase;
    OUTW(vgaIOBase + 4, (fg & 0x000000FF)<<8  | 0x48);
    OUTW(vgaIOBase + 4, (fg & 0x0000FF00)     | 0x49);
    OUTW(vgaIOBase + 4, (fg & 0x00FF0000)>>8  | 0x4A);
    OUTW(vgaIOBase + 4, (fg & 0xFF000000)>>16 | 0x4B);
    OUTW(vgaIOBase + 4, (bg & 0x000000FF)<<8  | 0x4C);
    OUTW(vgaIOBase + 4, (bg & 0x0000FF00)     | 0x4D);
    OUTW(vgaIOBase + 4, (bg & 0x00FF0000)>>8  | 0x4E);
    OUTW(vgaIOBase + 4, (bg & 0xFF000000)>>16 | 0x4F);
}

static void
TridentLoadCursorImage(
    ScrnInfoPtr pScrn,
    CARD8 *src
)
{
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    int vgaIOBase;
    vgaHWGetIOBase(VGAHWPTR(pScrn));
    vgaIOBase = VGAHWPTR(pScrn)->IOBase;

    memcpy((CARD8 *)pTrident->FbBase + (pScrn->videoRam * 1024) - 4096,
			src, pTrident->CursorInfoRec->MaxWidth * 
			pTrident->CursorInfoRec->MaxHeight / 4);

    OUTW(vgaIOBase + 4, (((pScrn->videoRam-4) & 0xFF) << 8) | 0x44);
    OUTW(vgaIOBase + 4, ((pScrn->videoRam-4) & 0xFF00) | 0x45);
}

static Bool 
TridentUseHWCursor(ScreenPtr pScreen, CursorPtr pCurs)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    
    if (pTrident->MUX && pScrn->bitsPerPixel == 8) return FALSE;

    if (!pTrident->HWCursor) return FALSE;

    return TRUE;
}

Bool 
TridentHWCursorInit(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    xf86CursorInfoPtr infoPtr;
    int memory = pScrn->displayWidth * pScrn->virtualY * pScrn->bitsPerPixel/8;

    if (memory > (pScrn->videoRam * 1024 - 4096)) return FALSE;
    infoPtr = xf86CreateCursorInfoRec();
    if(!infoPtr) return FALSE;
    
    pTrident->CursorInfoRec = infoPtr;

    infoPtr->MaxWidth = 64;
    infoPtr->MaxHeight = 64;
    infoPtr->Flags = HARDWARE_CURSOR_BIT_ORDER_MSBFIRST |
		HARDWARE_CURSOR_SWAP_SOURCE_AND_MASK |
		HARDWARE_CURSOR_SOURCE_MASK_INTERLEAVE_32 |
                ((pTrident->Chipset == CYBERBLADEE4) ? 
                HARDWARE_CURSOR_TRUECOLOR_AT_8BPP : 0);
    infoPtr->SetCursorColors = TridentSetCursorColors;
    infoPtr->SetCursorPosition = TridentSetCursorPosition;
    infoPtr->LoadCursorImage = TridentLoadCursorImage;
    infoPtr->HideCursor = TridentHideCursor;
    infoPtr->ShowCursor = TridentShowCursor;
    infoPtr->UseHWCursor = TridentUseHWCursor;

    return(xf86InitCursor(pScreen, infoPtr));
}

unsigned int
Tridentddc1Read(ScrnInfoPtr pScrn)
{
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    int vgaIOBase = VGAHWPTR(pScrn)->IOBase;
    CARD8 temp;

    /* New mode */
    OUTB(0x3C4, 0x0B); temp = INB(0x3C5);

    OUTB(0x3C4, NewMode1);
    temp = INB(0x3C5);
    OUTB(0x3C5, temp | 0x80);

    /* Define SDA as input */
    OUTW(vgaIOBase + 4, (0x04 << 8) | I2C);

    OUTW(0x3C4, (temp << 8) | NewMode1);

    /* Wait until vertical retrace is in progress. */
    while (INB(vgaIOBase + 0xA) & 0x08);
    while (!(INB(vgaIOBase + 0xA) & 0x08));

    /* Get the result */
    OUTB(vgaIOBase + 4, I2C);
    return ( INB(vgaIOBase + 5) & 0x01 );
}

void TridentSetOverscan(
    ScrnInfoPtr pScrn,
    int overscan
){
    TRIDENTPtr pTrident;
    vgaHWPtr hwp = VGAHWPTR(pScrn);
    unsigned char i,protect;
    unsigned char OScridx;

    pTrident = TRIDENTPTR(pScrn);
    if (overscan < 0 || overscan > 255)
	return;

    /* Add by ian ,not support PALM mode yet*/
    if(pTrident->TVChipset==1 && pTrident->TVOverscan==TRUE && pTrident->TVSignalMode<2){
    	if (pTrident->TVSignalMode == 0)
    	    OScridx=0x0;
    	else
    	    OScridx=0x1;
    	if (pScrn->depth==8 || pScrn->depth==16)
    	    OScridx=(OScridx << 1) | 0x0;
    	else
            OScridx=(OScridx << 1) | 0x1;
    	if (pScrn->currentMode->HDisplay==640 && pScrn->currentMode->VDisplay==480)
    	    OScridx=(OScridx << 1) | 0x0;
    	else
    	    OScridx=(OScridx << 1) | 0x1;

        /* Unprotect */
        outb(0x3C4,0x11);
        protect = inb(0x3C5);
        outb(0x3C5,0x92);
        /* set overscan */
        for (i=0; i<VT1621OSCRCRNUM; i++) {
            outb(0x3D4, VT1621OScrRegIndex.byCR[i]);
            outb(0x3D5, VT1621OScrRegValue[OScridx].byCR[i]);
        }
        for (i=0; i<VT1621OSCRSMNUM; i++) {
            smbus_write(VT1621OScrRegValue[OScridx].bySM[i], VT1621OScrRegIndex.bySM[i], TVX_VT1621_PORT);
        }
        /* protect */
        outb(0x3C4, 0x11);
        outb(0x3C5, protect);
    }

    hwp->enablePalette(hwp);
    hwp->writeAttr(hwp, OVERSCAN, overscan);
    hwp->disablePalette(hwp);
}

void TridentLoadPalette(
    ScrnInfoPtr pScrn, 
    int numColors, 
    int *indicies,
    LOCO *colors,
    VisualPtr pVisual
){
    vgaHWPtr hwp = VGAHWPTR(pScrn);
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    int i, index;
    for(i = 0; i < numColors; i++) {
	index = indicies[i];
    	OUTB(0x3C6, 0xFF);
	DACDelay(hwp);
        OUTB(0x3c8, index);
	DACDelay(hwp);
        OUTB(0x3c9, colors[index].red);
	DACDelay(hwp);
        OUTB(0x3c9, colors[index].green);
	DACDelay(hwp);
        OUTB(0x3c9, colors[index].blue);
	DACDelay(hwp);
    }
}

unsigned char smbus_read(unsigned char bIndex, unsigned char devAdr)
{
    unsigned short i;
    unsigned char  bData;

    /* clear host status */
    outb(SMBUS_BASE, 0xFF);

    /* check SMBUS ready */
    for ( i = 0; i < 0xFFFF; i++ )
        if ( (inb(SMBUS_BASE) & 0x01) == 0 )
            break;

    /* set host command */
    outb(SMBUS_BASE+3, bIndex);

    /* set slave address */
    outb(SMBUS_BASE+4, devAdr | 0x01);

    /* start */
    outb(SMBUS_BASE+2, 0x48);

    /* SMBUS Wait Ready */
    for ( i = 0; i < 0xFFFF; i++ )
        if ( (inb(SMBUS_BASE) & 0x01) == 0 )
            break;
    bData=inb(SMBUS_BASE+5);

    return bData;

}

void smbus_write(unsigned char bData, unsigned char bIndex, unsigned char devAdr)
{
    unsigned short i;

    /* clear host status */
    outb(SMBUS_BASE, 0xFF);

    /* check SMBUS ready */
    for ( i = 0; i < 0xFFFF; i++ )
        if ( (inb(SMBUS_BASE) & 0x01) == 0 )
           break;

    outb(SMBUS_BASE+2, 0x08);

    /* set host command */
    outb(SMBUS_BASE+3, bIndex);

    /* set slave address */
    outb(SMBUS_BASE+4, devAdr & 0xFE);

    outb(SMBUS_BASE+5, bData);

    /* start */
    outb(SMBUS_BASE+2, 0x48);

    /* SMBUS Wait Ready */
    for ( i = 0; i < 0xFFFF; i++ )
    if ( (inb(SMBUS_BASE) & 0x01) == 0 )
       break;
}
void VIA_SaveTVDepentVGAReg(ScrnInfoPtr pScrn)
{
    TRIDENTPtr pTrident=TRIDENTPTR(pScrn);
    unsigned char protect;
    unsigned char bTmp;
    int i;
    unsigned char VGA_RegIdx_about_TV[VGA_REGNUM_ABOUT_TV]={
                  0xD8,0XD9,/* SR */
                  0X33,/* GR */
                  0XC0,0XD0,0XD1,0XD2,0XD3,0XE0,0XE3,0XE4,0XE5,/* CR */
                  0XE6,0XE7,0XF0,0XF1,0XF6,0XFE,0XFF
                  };
    unsigned char TV_CH7005C_RegIdx[TV_CH7005C_TVREG_NUM]={
                  0X00,0X01,0X03,0X04,0X06,0X07,0X08,0X09,
		  0X0A,0X0B,0X0D,0X0E,0X10,0X11,0X13,0X14,
		  0X15,0X17,0X18,0X19,0X1A,0X1B,0X1C,0X1D,
		  0X1E,0X1F,0X20,0X21,0X3D
                  };

    /*ErrorF("VIAB3D: VIA_SaveTVDepentVGAReg:\n");*/

    /* Unprotect */
    outb(0x3C4, 0x11);
    protect = inb(0x3C5);
    outb(0x3C5, 0x92);

    /* Set TV Hw environment */
    outb(0x3d4,0xc1);
    outb(0x3d5,0x41);

    /* SR_d8,SR_d9 */
    for (i=0; i<2; i++)
    {
        outb(0x3c4,VGA_RegIdx_about_TV[i]);
        bTmp=inb(0x3c5);
        pTrident->DefaultTVDependVGASetting[i]=bTmp;
    }

    /* GR_33 */
    outb(0x3ce,0x33);
    bTmp=inb(0x3cf);
    pTrident->DefaultTVDependVGASetting[2]=bTmp;

    /* CR_c0,d0,d1,d2,d3,e0,e3,e4,e5,e6,e7,f0,f1,f6,fe,ff */
    for (i=3; i<VGA_REGNUM_ABOUT_TV; i++)
    {
        outb(0x3d4,VGA_RegIdx_about_TV[i]);
        bTmp=inb(0x3d5);
        pTrident->DefaultTVDependVGASetting[i]=bTmp;
    }

    switch (pTrident->TVChipset)
    {
       case 1:
             for (i=0; i<TVX_REG_NUM; i++)
             {
                 bTmp=smbus_read(i,TVX_VT1621_PORT);
                 pTrident->DefaultTVDependVGASetting[VGA_REGNUM_ABOUT_TV+i]=bTmp;
             }
	     break;
       case 2:
             for (i=0; i<TV_CH7005C_TVREG_NUM; i++)
	     {
                 bTmp=smbus_read(TV_CH7005C_RegIdx[i],TV_CH7005C_PORT);
                 pTrident->DefaultTVDependVGASetting[VGA_REGNUM_ABOUT_TV+i]=bTmp;
	     }
	     break;
       default:
             ErrorF("VIAB3D: VIA_SaveTVDepentVGAReg: Wrong Chipset setting\n");
	     break;

    }
    /* protect */
    outb(0x3C4, 0x11);
    outb(0x3C5, protect);
}
void VIA_RestoreTVDependVGAReg(ScrnInfoPtr pScrn)
{
    TRIDENTPtr pTrident=TRIDENTPTR(pScrn);
    unsigned char protect;
    unsigned char bTmp;
    int i;
    unsigned char VGA_RegIdx_about_TV[VGA_REGNUM_ABOUT_TV]={
                  0xD8,0XD9,/* SR */
                  0X33,/* GR */
                  0XC0,0XD0,0XD1,0XD2,0XD3,0XE0,0XE3,0XE4,0XE5,/* CR */
                  0XE6,0XE7,0XF0,0XF1,0XF6,0XFE,0XFF
                  };
    unsigned char TV_CH7005C_RegIdx[TV_CH7005C_TVREG_NUM]={
                  0X00,0X01,0X03,0X04,0X06,0X07,0X08,0X09,
		  0X0A,0X0B,0X0D,0X0E,0X10,0X11,0X13,0X14,
		  0X15,0X17,0X18,0X19,0X1A,0X1B,0X1C,0X1D,
		  0X1E,0X1F,0X20,0X21,0X3D
                  };

    /*ErrorF("VIAB3D: VIA_RestoreTVDependVGAReg:\n");*/

    /* Unprotect */
    outb(0x3C4, 0x11);
    protect = inb(0x3C5);
    outb(0x3C5, 0x92);

    /* Set TV Hw environment */
    outb(0x3d4,0xc1);
    outb(0x3d5,0x41);

    /* SR_d8,SR_d9 */
    for (i=0; i<2; i++)
    {
        outb(0x3c4,VGA_RegIdx_about_TV[i]);
        bTmp=pTrident->DefaultTVDependVGASetting[i];
        outb(0x3c5,bTmp);
    }
    /* GR_33 */
    outb(0x3ce,0x33);
    bTmp=pTrident->DefaultTVDependVGASetting[2];
    outb(0x3cf,bTmp);

    /* CR_c0,d0,d1,d2,d3,e0,e3,e4,e5,e6,e7,f0,f1,f6,fe,ff */
    for (i=3; i<VGA_REGNUM_ABOUT_TV; i++)
    {
        outb(0x3d4,VGA_RegIdx_about_TV[i]);
        bTmp=pTrident->DefaultTVDependVGASetting[i];
        outb(0x3d5,bTmp);
    }
    switch (pTrident->TVChipset)
    {
         case 1:
                for (i=0; i<TVX_REG_NUM; i++)
                {
                    bTmp=pTrident->DefaultTVDependVGASetting[VGA_REGNUM_ABOUT_TV+i];
                    smbus_write(bTmp,i,TVX_VT1621_PORT);
		}
		break;
         case 2:
	        for (i=0; i<TV_CH7005C_TVREG_NUM; i++)
		{
                    bTmp=pTrident->DefaultTVDependVGASetting[VGA_REGNUM_ABOUT_TV+i];
                    smbus_write(bTmp,TV_CH7005C_RegIdx[i],TV_CH7005C_PORT);
		}
		break;
       default:
             ErrorF("VIAB3D: VIA_SaveTVDepentVGAReg: Wrong Chipset setting\n");
	     break;
    }
    /* protect */
    outb(0x3C4, 0x11);
    outb(0x3C5, protect);
}
void VIA_TVInit(ScrnInfoPtr pScrn)
{
    TRIDENTPtr pTrident=TRIDENTPTR(pScrn);
    unsigned char idx=0;
    unsigned char i;
    unsigned char protect;
    unsigned char TV_CRTC[TVX_CRTC_NUM] =
       { 0xC0,0xD0,0xD1,0xD2,0xD3,0xE0,0xE3,0xE4,0xE5,
         0xE6,0xE7,0xF0,0xF1,0xF6,0xFE,0xFF  };
    unsigned char TV_CH7005C_RegIdx[TV_CH7005C_TVREG_NUM]={
                  0X00,0X01,0X03,0X04,0X06,0X07,0X08,0X09,
		  0X0A,0X0B,0X0D,0X0E,0X10,0X11,0X13,0X14,
		  0X15,0X17,0X18,0X19,0X1A,0X1B,0X1C,0X1D,
		  0X1E,0X1F,0X20,0X21,0X3D
                  };

#ifdef DEBUG_CODE_TRACE
    ErrorF("VIAB3D: VIA_TVInit:\n");
#endif

    if (pScrn->currentMode->HDisplay==640 && pScrn->currentMode->VDisplay==480 && (pScrn->depth==8 || pScrn->depth==16) && pTrident->TVSignalMode == 0)
    {
       /* Overlay window 1 position OK */
       ErrorF("VIAB3D: VIA_TVInit: TV Params 640x480x8(16) NTSC\n");
       idx=0;
       if(pTrident->TVOverscan==TRUE){
          pTrident->OverrideHsync+=-87;
          pTrident->OverrideVsync+=0;
       }else{
          pTrident->OverrideHsync+=-71;
          pTrident->OverrideVsync+=15;
       }
    }
    else if (pScrn->currentMode->HDisplay==800 && pScrn->currentMode->VDisplay==600 && (pScrn->depth==8 || pScrn->depth==16) && pTrident->TVSignalMode == 0)
    {
       /* Overlay window 1 position OK */
       ErrorF("VIAB3D: VIA_TVInit: TV Params 800x600x8(16) NTSC\n");      
       idx=1;
       if(pTrident->TVOverscan==TRUE){
          pTrident->OverrideHsync+=-95;
          pTrident->OverrideVsync+=4;
          pTrident->vsync_bskew+=1;
       }else{
          pTrident->OverrideHsync+=-152;
          pTrident->OverrideVsync+=72;
          pTrident->vsync_bskew+=1;
       }
    }
    else if (pScrn->currentMode->HDisplay==640 && pScrn->currentMode->VDisplay==480 && pScrn->depth==24 && pTrident->TVSignalMode == 0)
    {
       ErrorF("VIAB3D: VIA_TVInit: TV Params 640x480x24 NTSC\n");
       idx=2;
       if(pTrident->TVOverscan==TRUE){
          pTrident->OverrideHsync+=-88;
          pTrident->OverrideVsync+=0;
          pTrident->vsync_bskew+=1;
       }else{
          pTrident->OverrideHsync+=-72;
          pTrident->OverrideVsync+=14;
          pTrident->vsync_bskew+=1;
       }
    }
    else if (pScrn->currentMode->HDisplay==800 && pScrn->currentMode->VDisplay==600 && pScrn->depth==24 && pTrident->TVSignalMode == 0)
    {
       ErrorF("VIAB3D: VIA_TVInit: TV Params 800x600x24 NTSC\n");
       idx=3;
       if(pTrident->TVOverscan==TRUE){
          pTrident->OverrideHsync+=-96;
          pTrident->OverrideVsync+=2;
          pTrident->vsync_bskew+=1;
       }else{
          pTrident->OverrideHsync+=-160;
          pTrident->OverrideVsync+=72;
          pTrident->vsync_bskew+=1;
       }
    }
    else if (pScrn->currentMode->HDisplay==640 && pScrn->currentMode->VDisplay==480 && (pScrn->depth==8 || pScrn->depth==16) && pTrident->TVSignalMode == 1)
    {
       /* Overlay window 1 position OK */
       ErrorF("VIAB3D: VIA_TVInit: TV Params 640x480x8(16) PAL\n");
       idx=4;
       if(pTrident->TVOverscan==TRUE){
          pTrident->OverrideHsync+=-7;
          pTrident->OverrideVsync+=-18;
       }else{
          pTrident->OverrideHsync+=-7;
          pTrident->OverrideVsync+=65;
       }
    }
    else if (pScrn->currentMode->HDisplay==800 && pScrn->currentMode->VDisplay==600 && (pScrn->depth==8 || pScrn->depth==16) && pTrident->TVSignalMode == 1)
    {
       ErrorF("VIAB3D: VIA_TVInit: TV Params 800x600x8(16) PAL\n");
       /* patch TV screen defection */
       idx=5;
       /* patch 800x600 screen defect */
       outb(0x3d4,0x2f);
       outb(0x3d5,0xbf);
       if(pTrident->TVOverscan==TRUE){
          pTrident->OverrideHsync+=-151;
          pTrident->OverrideVsync+=-7;
       }else{
          pTrident->OverrideHsync+=-145;
          pTrident->OverrideVsync+=43;
       }
    }
    else if (pScrn->currentMode->HDisplay==640 && pScrn->currentMode->VDisplay==480 && pScrn->depth==24 && pTrident->TVSignalMode == 1)
    {
       ErrorF("VIAB3D: VIA_TVInit: TV Params 640x480x24 PAL\n");
       idx=6;
       if(pTrident->TVOverscan==TRUE){
          pTrident->OverrideHsync+=-8;
          pTrident->OverrideVsync+=-18;
          pTrident->vsync_bskew+=2;
       }else{
          pTrident->OverrideHsync+=-8;
          pTrident->OverrideVsync+=65;
          pTrident->vsync_bskew+=2;
       }
    }
    else if (pScrn->currentMode->HDisplay==800 && pScrn->currentMode->VDisplay==600 && pScrn->depth==24 && pTrident->TVSignalMode == 1)
    {
       ErrorF("VIAB3D: VIA_TVInit: TV Params 800x600x24 PAL\n");
       idx=7;
       outb(0x3d4,0x2f);
       outb(0x3d5,0xbf);
       if(pTrident->TVOverscan==TRUE){
          pTrident->OverrideHsync+=-152;
          pTrident->OverrideVsync+=-7;
          pTrident->vsync_bskew+=1;
       }else{
          pTrident->OverrideHsync+=-144;
          pTrident->OverrideVsync+=42;
          pTrident->vsync_bskew+=1;
       }
    }
    /* Add by ian */
    else if (pScrn->currentMode->HDisplay==640 && pScrn->currentMode->VDisplay==480 && (pScrn->depth==8 || pScrn->depth==16) && pTrident->TVSignalMode == 2)
    {
       /* Overlay window 1 position OK */
       ErrorF("VIAB3D: VIA_TVInit: TV Params 640x480x8(16) PALM\n");
       idx=8;
       pTrident->OverrideHsync+=-71;
       pTrident->OverrideVsync+=15;
       }
    else if (pScrn->currentMode->HDisplay==800 && pScrn->currentMode->VDisplay==600 && (pScrn->depth==8 || pScrn->depth==16) && pTrident->TVSignalMode == 2)
    {
       ErrorF("VIAB3D: VIA_TVInit: TV Params 800x600x8(16) PALM\n");
       /* patch TV screen defection */
       idx=9;
       /* patch 800x600 screen defect */
       outb(0x3d4,0x2f);
       outb(0x3d5,0xbf);
       pTrident->OverrideHsync+=-151;
       pTrident->OverrideVsync+=73;
       pTrident->vsync_bskew+=30;
       }
    else if (pScrn->currentMode->HDisplay==640 && pScrn->currentMode->VDisplay==480 && pScrn->depth==24 && pTrident->TVSignalMode == 2)
    {
       ErrorF("VIAB3D: VIA_TVInit: TV Params 640x480x24 PALM\n");
       idx=10;
       pTrident->OverrideHsync+=-73;
       pTrident->OverrideVsync+=15;
       }
    else if (pScrn->currentMode->HDisplay==800 && pScrn->currentMode->VDisplay==600 && pScrn->depth==24 && pTrident->TVSignalMode == 2)
    {
       ErrorF("VIAB3D: VIA_TVInit: TV Params 800x600x24 PALM\n");
       idx=11;
       outb(0x3d4,0x2f);
       outb(0x3d5,0xbf);
       pTrident->OverrideHsync+=-160;
       pTrident->OverrideVsync+=73;
       pTrident->vsync_bskew+=31;
       }
    else
    {
       ErrorF("VIAB3D: VIA_TVInit: TV Params default mode\n");
       return;
    }

    /* Unprotect */
    outb(0x3C4, 0x11);
    protect = inb(0x3C5);
    outb(0x3C5, 0x92);

    /* Set TV hw environment */
    outb(0x3c4,0x24);
    outb(0x3c5,0x4f);
    outb(0x3d4,0xc1);
    outb(0x3d5,0x41);
    outb(0x3ce,0x23);
    outb(0x3cf,0x88);

    /* set CRT + TV */
    outb(0x3CE,0x33);
    outb(0x3CF,0x20);

    /* set CRTC */
    for( i = 0; i < TVX_CRTC_NUM; i++ )
    {
         outb(0x3D4, TV_CRTC[i]);

	 if (pTrident->TVChipset==2)
            outb(0x3D5, TV_CH7005C_Table[idx][i]);
	 else outb(0x3D5, TVX_VT1621_Table[idx][i]);
    }


    /* Digital TV interface control */
    switch (pTrident->TVChipset)
    {
           case 1: outb(0x3C4,0xD8);
                   outb(0x3C5,0x60);
                   outb(0x3C4,0xD9);
                   outb(0x3C5,0x38);
		   break;
           case 2: outb(0x3c4,0xd8);
	           outb(0x3c5,0x24);
                   outb(0x3C4,0xD9);
                   outb(0x3C5,0x18);
		   break;
    }

    switch (pTrident->TVChipset)
    {
           case 1:
                  /* set TVX registers */
                  for (i=0; i < TVX_REG_NUM; i++ )
                  {
                      smbus_write(TVX_VT1621_Table[idx][TVX_CRTC_NUM+i], i, TVX_VT1621_PORT);
                  }
		  break;
           case 2:
	          for (i=0; i<TV_CH7005C_TVREG_NUM; i++)
		  {
                      smbus_write(TV_CH7005C_Table[idx][TV_CH7005C_CRTC_NUM+i], TV_CH7005C_RegIdx[i], TV_CH7005C_PORT);
		  }
	          break;
    }

    if (pTrident->TVRegSet)
    {      
       switch (pTrident->TVChipset)
       {
           case 1:
                  /* set TVX registers */
                  for (i=0; i < pTrident->TVRegCustNum; i++ )
                  {
                      smbus_write(pTrident->TVRegUserSet[1][i],pTrident->TVRegUserSet[0][i],TVX_VT1621_PORT);
                  }
		  break;
           case 2:
	          for (i=0; i < pTrident->TVRegCustNum; i++)
		  {
                      smbus_write(pTrident->TVRegUserSet[1][i],pTrident->TVRegUserSet[0][i], TV_CH7005C_PORT);
		  }
	          break;
       }
    }


    /*VIA_DumpReg(pScrn);*/

    /* protect */
    outb(0x3C4, 0x11);
    outb(0x3C5, protect);
}
void VIA_DumpReg(ScrnInfoPtr pScrn)
{
    TRIDENTPtr pTrident=TRIDENTPTR(pScrn);
    int i,j;
    unsigned char bTmp;
    unsigned char protect;

    /* Unprotect */
    outb(0x3C4, 0x11);
    protect = inb(0x3C5);
    outb(0x3C5, 0x92);

    /* SR */
    for (i=0; i<16; i++)
    {
        for (j=0; j<16; j++)
	{
            outb(0x3c4,(16*i+j));
	    bTmp=inb(0x3c5);

	    ErrorF("SR%02x=%02x ",(16*i+j),bTmp);
	}
	ErrorF("\n");
    }
    ErrorF("\n");
    /* CR */
    for (i=0; i<16; i++)
    {
        for (j=0; j<16; j++)
	{
            outb(0x3d4,(16*i+j));
	    bTmp=inb(0x3d5);

	    ErrorF("CR%02x=%02x ",(16*i+j),bTmp);
	}
	ErrorF("\n");
    }
    ErrorF("\n");
    /* GR */
    for (i=0; i<16; i++)
    {
        for (j=0; j<16; j++)
	{
            outb(0x3ce,(16*i+j));
	    bTmp=inb(0x3cf);

	    ErrorF("GR%02x=%02x ",(16*i+j),bTmp);
	}
	ErrorF("\n");
    }
    ErrorF("\n");
    /* SM */
    for (i=0; i<16; i++)
    {
        for (j=0; j<16; j++)
	{
	    if (pTrident->TVChipset==2)
               bTmp=smbus_read((16*i+j),TV_CH7005C_PORT);
	    else bTmp=smbus_read((16*i+j),TVX_VT1621_PORT);
	    ErrorF("SM%02x=%02x ",(16*i+j),bTmp);
	}
	ErrorF("\n");
    }
    ErrorF("\n");
    /* protect */
    outb(0x3C4, 0x11);
    outb(0x3C5, protect);

}
void VIA_PaserTVReg(ScrnInfoPtr pScrn,char* pTVReg)
{
    TRIDENTPtr pTrident=TRIDENTPTR(pScrn);
    int i;
    char bTmpH,bTmpL;
    int Num=0;

    pTrident->TVRegCustNum=0;

    /* Check input string num */
    while (*((char*)(pTVReg+Num)) != '\0')
    {
          Num++;
	  if (Num > 256)
	  break;
    }
    if (((Num+1)%3) != 0)
    {
       ErrorF("VIAB3D: VIA_PaserTVReg: Customered TV Reg format error\n");
       return;
    }

    pTrident->TVRegCustNum=(Num+1)/(3*2);
    ErrorF("pTrident->TVRegCustNum=%d\n",pTrident->TVRegCustNum);

    /* Check input string meet format? */
    for (i=0; i<2*pTrident->TVRegCustNum; i++)
    {
        bTmpH=*((char*)(pTVReg+3*i));
	bTmpL=*((char*)(pTVReg+3*i+1));
	if (i%2 == 0)
	{
	   if (bTmpH >= '0' && bTmpH <= '9')
	   {
              pTrident->TVRegUserSet[0][(int)(i/2)]=16*(bTmpH-'0');
	   }
	   else if (bTmpH >= 'a' && bTmpH <= 'f')
	   {
              pTrident->TVRegUserSet[0][(int)(i/2)]=16*(bTmpH-'a'+10);
	   }
	   else ErrorF("VIAB3D: VIA_PaserTVReg: TVReg paser error\n");

	   if (bTmpL >= '0' && bTmpL <= '9')
	   {
              pTrident->TVRegUserSet[0][(int)(i/2)]+=bTmpL-'0';
	   }
	   else if (bTmpL >= 'a' && bTmpL <= 'f')
	   {
              pTrident->TVRegUserSet[0][(int)(i/2)]+=bTmpL-'a'+10;
	   }
	   else ErrorF("VIAB3D: VIA_PaserTVReg: TVReg paser error\n");
	}
	else
	{
	   if (bTmpH >= '0' && bTmpH <= '9')
	   {
              pTrident->TVRegUserSet[1][(int)(i/2)]=16*(bTmpH-'0');
	   }
	   else if (bTmpH>= 'a' && bTmpH <= 'f')
	   {
              pTrident->TVRegUserSet[1][(int)(i/2)]=16*(bTmpH-'a'+10);
	   }
	   else ErrorF("VIAB3D: VIA_PaserTVReg: TVReg paser error\n");

	   if (bTmpL >= '0' && bTmpL <= '9')
	   {
              pTrident->TVRegUserSet[1][(int)(i/2)]+=bTmpL-'0';
	   }
	   else if (bTmpL >= 'a' && bTmpL <= 'f')
	   {
              pTrident->TVRegUserSet[1][(int)(i/2)]+=bTmpL-'a'+10;
	   }
	   else ErrorF("VIAB3D: VIA_PaserTVReg: TVReg paser error\n");
	}
    }
}

#if TRIDENT_VIP

void VIA_SaveVIPDepentVGAReg(ScrnInfoPtr pScrn)
{
    TRIDENTPtr pTrident=TRIDENTPTR(pScrn);
    unsigned char protect;
    unsigned char bTmp;
    int i;    

#if Video_DEBUG
    ErrorF("VIAB3D: VIA_SaveVIPDepentVGAReg:\n");
#endif

    /* Unprotect */
    outb(0x3C4, 0x11);
    protect = inb(0x3C5);
    outb(0x3C5, 0x92);

    /* CR 0x80-0xbf */
    for (i=0; i<64; i++)
    {
        outb(0x3d4,0x80+i);
        bTmp=inb(0x3d5);
        pTrident->DefaultVIPDependVGA_CR_Setting[i]=bTmp;
    }
    /* SR 0x90-0xbf */
    for (i=0; i<48; i++)
    {
        outb(0x3c4,0x90+i);
        bTmp=inb(0x3c5);
        pTrident->DefaultVIPDependVGA_SR_Setting[i]=bTmp;
    }

    /* protect */
    outb(0x3C4, 0x11);
    outb(0x3C5, protect);
}
void VIA_RestoreVIPDependVGAReg(ScrnInfoPtr pScrn)
{
    TRIDENTPtr pTrident=TRIDENTPTR(pScrn);
    unsigned char protect;
    unsigned char bTmp;
    unsigned char ucTemp;
    int i;
    
#if Video_DEBUG
    ErrorF("VIAB3D: VIA_RestoreVIPDependVGAReg:\n");
#endif 

    /* Unprotect */
    outb(0x3C4, 0x11);
    protect = inb(0x3C5);
    outb(0x3C5, 0x92);

    /* Disable video capture */
  	ucTemp =  getCR(0xaf);
    ucTemp &= 0x7f; 
  	setCR(0xaf, ucTemp);
  
  	ucTemp =  getCR(0x9c);
    ucTemp &= 0xef; 
  	setCR(0x9c, ucTemp);
	
    /* CR 0x80-0xbf */
    for (i=0; i<64; i++)
    {
        outb(0x3d4,0x80+i);
        bTmp=pTrident->DefaultVIPDependVGA_CR_Setting[i];
        outb(0x3d5,bTmp);
    }
    /* SR 0x90-0xbf */
    for (i=0; i<48; i++)
    {
        outb(0x3c4,0x90+i);
        bTmp=pTrident->DefaultVIPDependVGA_SR_Setting[i];
        outb(0x3c5,bTmp);
    }

    /* protect */
    outb(0x3C4, 0x11);
    outb(0x3C5, protect);
}
#endif
