/*
 * Copyright 1992-2000 by Alan Hourihane, Wigan, England.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Alan Hourihane not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Alan Hourihane makes no representations
 * about the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 * ALAN HOURIHANE DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL ALAN HOURIHANE BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Alan Hourihane, alanh@fairlite.demon.co.uk
 */
/* $XFree86: xc/programs/Xserver/hw/xfree86/drivers/trident/trident_video.c,v 1.20 2001/12/13 18:01:51 eich Exp $ */

#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86Resources.h"
#include "xf86_ansic.h"
#include "compiler.h"
#include "xf86PciInfo.h"
#include "xf86Pci.h"
#include "xf86fbman.h"
#include "regionstr.h"

#include "trident.h"
#include "trident_regs.h"
#include "Xv.h"
#include "xaa.h"
#include "xaalocal.h"
#include "dix.h"
#include "dixstruct.h"
#include "fourcc.h"

#if TRIDENT_VIP
#include "xf86PciInfo.h"
#include "xf86i2c.h"
/*#include "generic_bus.h"*/
#endif 

#define OFF_DELAY 	200  /* milliseconds */
#define FREE_DELAY 	60000

#define OFF_TIMER 	0x01
#define FREE_TIMER	0x02
#define CLIENT_VIDEO_ON	0x04

#define TIMER_MASK      (OFF_TIMER | FREE_TIMER)

static XF86VideoAdaptorPtr TRIDENTSetupImageVideo(ScreenPtr);
static void TRIDENTInitOffscreenImages(ScreenPtr);
static void TRIDENTStopVideo(ScrnInfoPtr, pointer, Bool);
static int TRIDENTSetPortAttribute(ScrnInfoPtr, Atom, INT32, pointer);
static int TRIDENTGetPortAttribute(ScrnInfoPtr, Atom ,INT32 *, pointer);
static void TRIDENTQueryBestSize(ScrnInfoPtr, Bool,
	short, short, short, short, unsigned int *, unsigned int *, pointer);
static int TRIDENTPutImage( ScrnInfoPtr, 
	short, short, short, short, short, short, short, short,
	int, unsigned char*, short, short, Bool, RegionPtr, pointer);
static int TRIDENTQueryImageAttributes(ScrnInfoPtr, 
	int, unsigned short *, unsigned short *,  int *, int *);
static void TRIDENTVideoTimerCallback(ScrnInfoPtr pScrn, Time time);
static void tridentSetVideoGamma(TRIDENTPtr pTrident,int value,int brightness);
static void tridentSetVideoContrast(TRIDENTPtr pTrident,int value);
static void tridentSetVideoParameters(TRIDENTPtr pTrident, int brightness, 
				      int saturation, int hue);
void tridentFixFrame(ScrnInfoPtr pScrn, int *fixFrame);

#if TRIDENT_VIP
static int TRIDENTPutVideo( ScrnInfoPtr, 
	short, short, short, short, short, short, short, short,
	RegionPtr, pointer);

void ResetCapturePort(void);
void InitializeCapturePort(void);
void SetMinifyFactor(short src_w, short src_h, short drw_w, short drw_h);
void CapturePortSetting(ScrnInfoPtr pScrn, int offset);
void EnableCapturePort(void);
void DisableCapturePort(void);

unsigned char getCR(unsigned char nIndex);
unsigned char getSR(unsigned char nIndex);
void setCR(unsigned char nIndex, unsigned char nValue);
void setSR(unsigned char nIndex, unsigned char nValue);
int bVIP_ON = 0;
int bVIP_Init = 0;
#endif

#define MAKE_ATOM(a) MakeAtom(a, sizeof(a) - 1, TRUE)

#if TRIDENT_VIP

   static Atom xvColorKey, xvBrightness, xvContrast, xvSaturation, xvHue, xvGamma,
               xvColor, xvDoubleBuffer, xvEncoding, xvVolume, xvMute, 
	             xvFrequency, xv_autopaint_colorkey, xv_set_defaults, 
	             xvTunerStatus;

#else

   static Atom xvColorKey, xvSaturation, xvBrightness, xvHue, xvGamma, xvContrast;
#endif

void TRIDENTInitVideo(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    XF86VideoAdaptorPtr *adaptors, *newAdaptors = NULL;
    XF86VideoAdaptorPtr newAdaptor = NULL;
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    int num_adaptors;

#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTInitVideo:\n");
#endif
    /*
     * The following has been tested on:
     *
     * 9525         : flags: None
     * CyberBlade/i7: flags: VID_ZOOM_INV | VID_ZOOM_MINI
     * CyberBlade/i1: flags: VID_ZOOM_INV | VID_ZOOM_MINI
     * CyberBlade/Ai1: flags: VID_ZOOM_INV
     * Cyber 9540   : flags: VID_ZOOM_INV | VID_SHIFT_4
     * CyberXPm8    : flags: VID_ZOOM_INV | VID_SHIFT_4
     *
     * When you make changes make sure not to break these
     * Add new chipsets to this list.
     */
    if (pTrident->Chipset >= BLADE3D) {
       pTrident->videoFlags = VID_ZOOM_INV ;
       if (pTrident->Chipset <= CYBERBLADEI1D)
	 pTrident->videoFlags |= VID_ZOOM_MINI;
       else if (pTrident->Chipset < CYBERBLADEAI1 /* verified EE */
		|| pTrident->Chipset > CYBERBLADEAI1D)
	 pTrident->videoFlags |= VID_OFF_SHIFT_4;
    }

    newAdaptor = TRIDENTSetupImageVideo(pScreen);
    TRIDENTInitOffscreenImages(pScreen);

    num_adaptors = xf86XVListGenericAdaptors(pScrn, &adaptors);

    if(newAdaptor) {
	if(!num_adaptors) {
	    num_adaptors = 1;
	    adaptors = &newAdaptor;
	} else {
	    newAdaptors =  /* need to free this someplace */
		xalloc((num_adaptors + 1) * sizeof(XF86VideoAdaptorPtr*));
	    if(newAdaptors) {
		memcpy(newAdaptors, adaptors, num_adaptors * 
					sizeof(XF86VideoAdaptorPtr));
		newAdaptors[num_adaptors] = newAdaptor;
		adaptors = newAdaptors;
		num_adaptors++;
	    }
	}
    }

    if(num_adaptors)
        xf86XVScreenInit(pScreen, adaptors, num_adaptors);

    if(newAdaptors)
	xfree(newAdaptors);

    xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,3,"XvFlags: %s %s %s\n",
		   pTrident->videoFlags & VID_ZOOM_INV ? "VID_ZOOM_INV" : "",
		   pTrident->videoFlags & VID_ZOOM_MINI ? "VID_ZOOM_MINI" : "",
		   pTrident->videoFlags & VID_OFF_SHIFT_4 ? "VID_OFF_SHIFT_4"
		   : "");
		   
}

/* client libraries expect an encoding */
static XF86VideoEncodingRec DummyEncoding[1] =
{
 {
   0,
   "XV_IMAGE",
   1024, 1024,
   {1, 1}
 }
};

#if TRIDENT_VIP

static XF86VideoEncodingRec InputVideoEncodings[] =
{
  {0,"XV_IMAGE",		1024,1024,{1,1}},
  {1,"pal-composite",		720, 288, {1,50}},
  {2,"pal-tuner",		720, 288, {1,50}},
  {3,"pal-svideo",		720, 288, {1,50}},
  {4,"ntsc-composite",		720, 240, {1001,60000}},
  {5,"ntsc-tuner",		720, 240, {1001,60000}},
  {6,"ntsc-svideo",		720, 240, {1001,60000}},
  {7,"secam-composite", 	720, 240, {1,50}},
  {8,"secam-tuner",		720, 240, {1,50}},
  {9,"secam-svideo",		720, 240, {1,50}},
  {10,"pal_60-composite", 	768, 288, {1,50}},
  {11,"pal_60-tuner", 		768, 288, {1,50}},
  {12,"pal_60-svideo", 		768, 288, {1,50}},
};

#endif

#define NUM_FORMATS 4

static XF86VideoFormatRec Formats[NUM_FORMATS] =
{
  {8, PseudoColor},  {15, TrueColor}, {16, TrueColor}, {24, TrueColor}
};

#if TRIDENT_VIP

#define NUM_ATTRIBUTES 14

static XF86AttributeRec Attributes[NUM_ATTRIBUTES+1] =
{
   {XvSettable		   , 0, 1, "XV_SET_DEFAULTS"},
   {XvSettable | XvGettable, 0, (1 << 24) - 1, "XV_COLORKEY"},
   {XvSettable | XvGettable, -128, 127, "XV_BRIGHTNESS"},
   {XvSettable | XvGettable, 0, 255, "XV_CONTRAST"},
   {XvSettable | XvGettable, 0, 255, "XV_SATURATION"},
   {XvSettable | XvGettable, -180, 180, "XV_HUE"},
   {XvSettable | XvGettable, 0, 1,  "XV_DOUBLE_BUFFER"},
   {XvSettable | XvGettable, 0, 1,  "XV_AUTOPAINT_COLORKEY"},
   {XvSettable | XvGettable, 0, 12,  "XV_ENCODING"},
   {XvSettable | XvGettable, 0, -1,  "XV_FREQ"},
   {XvSettable | XvGettable, -1000, 1000,  "XV_TUNER_STATUS"},
   {XvSettable | XvGettable, 0, 255,  "XV_COLOR"},
   {XvSettable | XvGettable, 0, 1,  "XV_MUTE"},
   {XvSettable | XvGettable, 0x01, 0x7F,  "XV_VOLUME"},
   {0,0,0,NULL} 
};


#else

#define NUM_ATTRIBUTES 6

static XF86AttributeRec Attributes[NUM_ATTRIBUTES] =
{
    {XvSettable | XvGettable, 0, (1 << 24) - 1, "XV_COLORKEY"},
    {XvSettable | XvGettable, 0, 187,           "XV_SATURATION"},
    {XvSettable | XvGettable, 0, 0x3F,          "XV_BRIGHTNESS"},
    {XvSettable | XvGettable, 0, 360 ,          "XV_HUE"},
    {XvSettable | XvGettable, -128, 127,        "XV_GAMMA"},
    {XvSettable | XvGettable, 0, 7,           "XV_CONTRAST"}
};

#endif

#define NUM_IMAGES 4

static XF86ImageRec Images[NUM_IMAGES] =
{
   {
	0x35315652,
        XvRGB,
	LSBFirst,
	{'R','V','1','5',
	  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
	16,
	XvPacked,
	1,
	15, 0x001F, 0x03E0, 0x7C00,
	0, 0, 0,
	0, 0, 0,
	0, 0, 0,
	{'R','V','B',0,
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	XvTopToBottom
   },
   {
	0x36315652,
        XvRGB,
	LSBFirst,
	{'R','V','1','6',
	  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
	16,
	XvPacked,
	1,
	16, 0x001F, 0x07E0, 0xF800,
	0, 0, 0,
	0, 0, 0,
	0, 0, 0,
	{'R','V','B',0,
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	XvTopToBottom
   },
   XVIMAGE_YV12,
   XVIMAGE_YUY2
};

typedef struct {
   
#if TRIDENT_VIP
   Bool		doubleBuffer;
   unsigned char currentBuffer;
   int 		encoding;
   CARD32	frequency;
   int		volume;
   int		mute;
   Bool		autopaint_colorkey;
#endif

   FBLinearPtr	linear;
   RegionRec	clip;
   CARD32	colorKey;
   CARD8        Saturation;
   CARD8        Brightness;
   CARD16       HUE;
   INT8         Gamma;
   INT8         Contrast;
   CARD32	videoStatus;
   Time		offTime;
   Time		freeTime;
   int          fixFrame;
} TRIDENTPortPrivRec, *TRIDENTPortPrivPtr;


#define GET_PORT_PRIVATE(pScrn) \
   (TRIDENTPortPrivPtr)((TRIDENTPTR(pScrn))->adaptor->pPortPrivates[0].ptr)

void TRIDENTResetVideo(ScrnInfoPtr pScrn) 
{
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    TRIDENTPortPrivPtr pPriv = pTrident->adaptor->pPortPrivates[0].ptr;
    int vgaIOBase = VGAHWPTR(pScrn)->IOBase;
    int red, green, blue;
    int tmp;

#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTResetVideo:\n");
#endif

#if TRIDENT_VIP

    xvBrightness   = MAKE_ATOM("XV_BRIGHTNESS");
    xvSaturation   = MAKE_ATOM("XV_SATURATION");
    xvColor        = MAKE_ATOM("XV_COLOR");
    xvContrast     = MAKE_ATOM("XV_CONTRAST");
    xvColorKey     = MAKE_ATOM("XV_COLORKEY");
    xvDoubleBuffer = MAKE_ATOM("XV_DOUBLE_BUFFER");
    xvEncoding     = MAKE_ATOM("XV_ENCODING");
    xvTunerStatus  = MAKE_ATOM("XV_TUNER_STATUS");
    xvFrequency    = MAKE_ATOM("XV_FREQ");
    xvVolume       = MAKE_ATOM("XV_VOLUME");
    xvMute         = MAKE_ATOM("XV_MUTE");
    xvHue          = MAKE_ATOM("XV_HUE");
    xv_autopaint_colorkey = MAKE_ATOM("XV_AUTOPAINT_COLORKEY");
    xv_set_defaults = MAKE_ATOM("XV_SET_DEFAULTS");
#endif

    OUTW(vgaIOBase + 4, 0x008E);
    OUTW(vgaIOBase + 4, 0x008F);

    if (pTrident->Chipset >= CYBER9397) {
    	OUTW(vgaIOBase + 4, 0x80B9);
    	OUTW(0x3C4, 0xC057);
    	OUTW(0x3C4, 0x3420); /* Ian: For the right garbage in the playback */
    	OUTW(0x3C4, 0x3037);
    } else {
	if (pTrident->Chipset >= PROVIDIA9682) {
    	    OUTB(0x83C8, 0x57);
    	    OUTB(0x83C6, 0xC0);
    	    OUTW(vgaIOBase + 4, 0x24BE);
	} else {
    	    OUTB(0x83C8, 0x37);
    	    OUTB(0x83C6, 0x01);
	}
    }

    if (pTrident->Chipset >= CYBERBLADEXPm8) {
	OUTW(0x3C4, 0x007A);
	OUTW(0x3C4, 0x007D);
    }

    switch (pScrn->depth) {
    case 8:
	VIDEOOUT(pPriv->colorKey, pTrident->keyOffset);
	VIDEOOUT(0x00, (pTrident->keyOffset + 1));
	VIDEOOUT(0x00, (pTrident->keyOffset + 2));
	VIDEOOUT(0xFF, (pTrident->keyOffset + 4));
	VIDEOOUT(0x00, (pTrident->keyOffset + 5));
	VIDEOOUT(0x00, (pTrident->keyOffset + 6));
	break;
    default:
	red = (pPriv->colorKey & pScrn->mask.red) >> pScrn->offset.red;
	green = (pPriv->colorKey & pScrn->mask.green) >> pScrn->offset.green;
	blue = (pPriv->colorKey & pScrn->mask.blue) >> pScrn->offset.blue;
	switch (pScrn->depth) {
	case 15:
	    tmp = (red << 10) | (green << 5) | (blue);
	    VIDEOOUT((tmp & 0xff), pTrident->keyOffset);
	    VIDEOOUT((tmp & 0xff00)>>8, (pTrident->keyOffset + 1));
	    VIDEOOUT(0x00, (pTrident->keyOffset + 2));
	    VIDEOOUT(0xFF, (pTrident->keyOffset + 4));
	    VIDEOOUT(0xFF, (pTrident->keyOffset + 5));
	    VIDEOOUT(0x00, (pTrident->keyOffset + 6));
	    break;
	case 16:
	    tmp = (red << 11) | (green << 5) | (blue);
	    VIDEOOUT((tmp & 0xff), pTrident->keyOffset);
	    VIDEOOUT((tmp & 0xff00)>>8, (pTrident->keyOffset + 1));
	    VIDEOOUT(0x00, (pTrident->keyOffset + 2));
	    VIDEOOUT(0xFF, (pTrident->keyOffset + 4));
	    VIDEOOUT(0xFF, (pTrident->keyOffset + 5));
	    VIDEOOUT(0x00, (pTrident->keyOffset + 6));
	    break;
	case 24:
	    VIDEOOUT(blue, pTrident->keyOffset);
	    VIDEOOUT(green, (pTrident->keyOffset + 1));
	    VIDEOOUT(red, (pTrident->keyOffset + 2));
	    VIDEOOUT(0xFF, (pTrident->keyOffset + 4));
	    VIDEOOUT(0xFF, (pTrident->keyOffset + 5));
	    VIDEOOUT(0xFF, (pTrident->keyOffset + 6));
	    break;
	}    
    }    
    tridentSetVideoGamma(pTrident,pPriv->Gamma,pPriv->Brightness);
    tridentSetVideoContrast(pTrident,pPriv->Contrast);
    tridentSetVideoParameters(pTrident,pPriv->Brightness,pPriv->Saturation,
                            pPriv->HUE);
}


static XF86VideoAdaptorPtr 
TRIDENTSetupImageVideo(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    XF86VideoAdaptorPtr adapt;
    TRIDENTPortPrivPtr pPriv;

#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTSetupImageVideo:\n");
#endif

    if(!(adapt = xcalloc(1, sizeof(XF86VideoAdaptorRec) +
			    sizeof(TRIDENTPortPrivRec) +
			    sizeof(DevUnion))))
	return NULL;

#if TRIDENT_VIP
    adapt->type		= XvWindowMask | XvInputMask | XvImageMask |
                    XvVideoMask;
#else
    adapt->type = XvWindowMask | XvInputMask | XvImageMask;
#endif
    
    adapt->flags = VIDEO_OVERLAID_IMAGES | VIDEO_CLIP_TO_VIEWPORT;
    adapt->name = "Trident Backend Scaler";

#if TRIDENT_VIP
    adapt->nEncodings 		= 13;
    adapt->pEncodings 		= InputVideoEncodings;

#else
    adapt->nEncodings = 1;
    adapt->pEncodings = DummyEncoding;
#endif
    
    adapt->nFormats = NUM_FORMATS;
    adapt->pFormats = Formats;
    adapt->nPorts = 1;
    adapt->pPortPrivates = (DevUnion*)(&adapt[1]);
    pPriv = (TRIDENTPortPrivPtr)(&adapt->pPortPrivates[1]);
    adapt->pPortPrivates[0].ptr = (pointer)(pPriv);
    adapt->pAttributes = Attributes;
    adapt->nImages = NUM_IMAGES;
    adapt->nAttributes = NUM_ATTRIBUTES;
    adapt->pImages = Images;
    
#if TRIDENT_VIP
    adapt->PutVideo		= TRIDENTPutVideo;
#else
    adapt->PutVideo = NULL;
#endif

    adapt->PutStill = NULL;
    adapt->GetVideo = NULL;
    adapt->GetStill = NULL;
    adapt->StopVideo = TRIDENTStopVideo;
    adapt->SetPortAttribute = TRIDENTSetPortAttribute;
    adapt->GetPortAttribute = TRIDENTGetPortAttribute;
    adapt->QueryBestSize = TRIDENTQueryBestSize;
    adapt->PutImage = TRIDENTPutImage;
    adapt->QueryImageAttributes = TRIDENTQueryImageAttributes;

    pPriv->colorKey = pTrident->videoKey & ((1 << pScrn->depth) - 1);
    pPriv->Brightness = 45;
    pPriv->Saturation = 80;
    pPriv->Contrast = 4;
    pPriv->Gamma = 0;
    pPriv->HUE = 0;
    pPriv->videoStatus = 0;
    pPriv->fixFrame = 100;

    /* gotta uninit this someplace */
    REGION_INIT(pScreen, &pPriv->clip, NullBox, 0); 

    pTrident->adaptor = adapt;

    xvColorKey   = MAKE_ATOM("XV_COLORKEY");
   
    if (pTrident->Chipset >= CYBER9397) {
    	xvBrightness = MAKE_ATOM("XV_BRIGHTNESS");
    	xvSaturation = MAKE_ATOM("XV_SATURATION");
    	xvHue        = MAKE_ATOM("XV_HUE");
    	xvGamma      = MAKE_ATOM("XV_GAMMA");
    	xvContrast   = MAKE_ATOM("XV_CONTRAST");
    }

    if (pTrident->Chipset >= PROVIDIA9682) 
	pTrident->keyOffset = 0x50;
    else
	pTrident->keyOffset = 0x30;

    TRIDENTResetVideo(pScrn);

    return adapt;
}


static Bool
RegionsEqual(RegionPtr A, RegionPtr B)
{
    int *dataA, *dataB;
    int num;

    num = REGION_NUM_RECTS(A);
    if(num != REGION_NUM_RECTS(B))
	return FALSE;

    if((A->extents.x1 != B->extents.x1) ||
       (A->extents.x2 != B->extents.x2) ||
       (A->extents.y1 != B->extents.y1) ||
       (A->extents.y2 != B->extents.y2))
	return FALSE;

    dataA = (int*)REGION_RECTS(A);
    dataB = (int*)REGION_RECTS(B);

    while(num--) {
	if((dataA[0] != dataB[0]) || (dataA[1] != dataB[1]))
	   return FALSE;
	dataA += 2; 
	dataB += 2;
    }

    return TRUE;
}

#define DummyScreen screenInfo.screens[0]

static void 
TRIDENTStopVideo(ScrnInfoPtr pScrn, pointer data, Bool shutdown)
{
  TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
  TRIDENTPortPrivPtr pPriv = (TRIDENTPortPrivPtr)data;
  int vgaIOBase = VGAHWPTR(pScrn)->IOBase;

#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTStopVideo:\n");
#endif

  REGION_EMPTY(pScrn->pScreen, &pPriv->clip);

  if(shutdown) {
     if(pPriv->videoStatus & CLIENT_VIDEO_ON) {
	OUTW(vgaIOBase + 4, 0x008E);
	OUTW(vgaIOBase + 4, 0x008F);
     }
     if(pPriv->linear) {
	xf86FreeOffscreenLinear(pPriv->linear);
	pPriv->linear = NULL;
     }
     pPriv->videoStatus = 0;
  } else {
     if(pPriv->videoStatus & CLIENT_VIDEO_ON) {
	pPriv->videoStatus |= OFF_TIMER;
	pPriv->offTime = currentTime.milliseconds + OFF_DELAY;
	pTrident->VideoTimerCallback = TRIDENTVideoTimerCallback;
     }
  }
#if TRIDENT_VIP
  if (bVIP_ON)
  {
    DisableCapturePort();
    bVIP_ON = 0;
  }
#endif
}

#undef PI
#define PI 3.14159265

static void
tridentSetVideoContrast(TRIDENTPtr pTrident,int value)
{
#if Video_DEBUG
    ErrorF("VIAB3D: tridentSetVideoContrast:\n");
#endif
  OUTW(0x3C4, (((value & 0x7)|((value & 0x7) << 4)) << 8) | 0xBC);
}

static void
tridentSetVideoGamma(TRIDENTPtr pTrident,int value,int brightness)
{
  int pivots[] = {0,3,15,63,255};

  int slope;
  int y_0;
  float x, x_prev = 0, y, y_prev = 0;
  int i;
  int exp;
  CARD8 i_slopes[4];
  CARD8 intercepts[4];

#if Video_DEBUG
    ErrorF("VIAB3D: tridentSetVideoGamma:\n");
#endif

  brightness = (brightness - 0x20);
  if (value == 0) {
      OUTW(0x3C4, 0x80 << 8 | 0xB4);
      OUTW(0x3C4, (brightness) << 8 | 0xB8);
      OUTW(0x3C4, (brightness) << 8 | 0xB9);
      OUTW(0x3C4, (brightness) << 8 | 0xBA);
      OUTW(0x3C4, (brightness) << 8 | 0xBB);
      return;
  }
  exp = log(value);
  for (i = 0; i < 4; i++) {
    x = pivots[i-1] / 255.0;
    y = pow(x,exp);
    slope = (y - y_prev) / (x - x_prev);
    y_0 = y - x * slope;
    {
      int val = slope;
	  if (val > 7)
	    i_slopes[i] = (3 << 4) | (val & 0xf);
	  else if (val > 3)
	  i_slopes[i] = (2 << 4) | ((int)(slope / 2) & 0xf);
	  else if (val > 2)
	  i_slopes[i] = (1 << 4) | ((int)(slope / 4) & 0xf);
	  else
	    i_slopes[i] = ((int)(slope / 8) & 0xf);
    }
    intercepts[i] = y_0 * 256 / 4;
    x_prev = x;
    y_prev = y;
  }
      OUTW(0x3C4, i_slopes[0] << 8 | 0xB4);
      OUTW(0x3C4, i_slopes[1] << 8 | 0xB5);
      OUTW(0x3C4, i_slopes[2] << 8 | 0xB6);
      OUTW(0x3C4, i_slopes[3] << 8 | 0xB7);
      OUTW(0x3C4, (intercepts[0] + brightness) << 8 | 0xB8);
      OUTW(0x3C4, (intercepts[1] + brightness) << 8 | 0xB9);
      OUTW(0x3C4, (intercepts[2] + brightness) << 8 | 0xBA);
      OUTW(0x3C4, (intercepts[3] + brightness) << 8 | 0xBB);
}

static void
tridentSetVideoParameters(TRIDENTPtr pTrident, int brightness,
			  int saturation, int hue)
{
    double dtmp;
    CARD8 sign, tmp, tmp1;

#if Video_DEBUG
    ErrorF("VIAB3D: tridentSetVideoParamsters:\n");
#endif

    if (brightness >= 0x20)
      brightness -= 0x20;
    else
      brightness += 0x20;
    dtmp = sin((double)hue / 180.0 * PI) * saturation / 12.5;
    sign = (dtmp < 0) ? 1 << 1 : 0;
    tmp1 = ((int)fabs(dtmp) >> 4) & 0x1;
    tmp = brightness << 2 | sign | tmp1;
    OUTW(0x3C4, tmp << 8 | 0xB1);

    tmp1 = ((int)fabs(dtmp) & 0x7 ) << 5;
    dtmp = cos((double)hue / 180.0 * PI) * saturation / 12.5;
    sign = (dtmp < 0) ? 1 << 4 : 0;
    tmp1 |= (int)fabs(dtmp)  & 0xf;
    tmp = sign | tmp1;
    OUTW(0x3C4, tmp << 8 | 0xB0);
}

static int 
TRIDENTSetPortAttribute(
  ScrnInfoPtr pScrn, 
  Atom attribute,
  INT32 value, 
  pointer data
){
  TRIDENTPortPrivPtr pPriv = (TRIDENTPortPrivPtr)data;
  TRIDENTPtr pTrident = TRIDENTPTR(pScrn);

#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTSetPortAttribute:\n");
#endif
  
  if(attribute == xvColorKey) {
	int red, green, blue;
	int tmp;
	pPriv->colorKey = value;
	switch (pScrn->depth) {
	case 8:
	    VIDEOOUT(pPriv->colorKey, pTrident->keyOffset);
	    VIDEOOUT(0x00, (pTrident->keyOffset + 1));
	    VIDEOOUT(0x00, (pTrident->keyOffset + 2));
	    break;
	default:
	    red = (pPriv->colorKey & pScrn->mask.red) >> pScrn->offset.red;
	    green = (pPriv->colorKey & pScrn->mask.green) >> pScrn->offset.green;
	    blue = (pPriv->colorKey & pScrn->mask.blue) >> pScrn->offset.blue;
	    switch (pScrn->depth) {
	    case 15:
	    	tmp = (red << 10) | (green << 5) | (blue);
	    	VIDEOOUT((tmp&0xff), pTrident->keyOffset);
	    	VIDEOOUT((tmp&0xff00)>>8, (pTrident->keyOffset + 1));
	    	VIDEOOUT(0x00, (pTrident->keyOffset + 2));
		break;
	    case 16:
	    	tmp = (red << 11) | (green << 5) | (blue);
	    	VIDEOOUT((tmp&0xff), pTrident->keyOffset);
	    	VIDEOOUT((tmp&0xff00)>>8, (pTrident->keyOffset + 1));
	    	VIDEOOUT(0x00, (pTrident->keyOffset + 2));
		break;
	    case 24:
	    	VIDEOOUT(blue, pTrident->keyOffset);
	    	VIDEOOUT(green, (pTrident->keyOffset + 1));
	    	VIDEOOUT(red, (pTrident->keyOffset + 2));
		break;
	    }    
	}    /* Louis: Disable value check for xine -V xv */
	REGION_EMPTY(pScrn->pScreen, &pPriv->clip);   
  } else if (attribute == xvBrightness) {
    /*if ((value < 0) || (value > 0x3f))
      return BadValue;*/
    pPriv->Brightness = value;
    tridentSetVideoParameters(pTrident, pPriv->Brightness, pPriv->Saturation,
			      pPriv->HUE);
  } else if (attribute == xvSaturation) {
    /*if ((value < 0) || (value > 187))
      return BadValue;*/
    pPriv->Saturation = value;
    tridentSetVideoParameters(pTrident, pPriv->Brightness, pPriv->Saturation,
			      pPriv->HUE);
  } else if (attribute == xvHue) {
    /*if ((value < 0) || (value > 360))
      return BadValue;*/
    pPriv->HUE = value;
    tridentSetVideoParameters(pTrident, pPriv->Brightness, pPriv->Saturation,
			      pPriv->HUE);
    tridentSetVideoGamma(pTrident,pPriv->Gamma,pPriv->Brightness);
  } else if (attribute == xvGamma) {
    /*if ((value < -128) || (value > 127))
      return BadValue;*/
    pPriv->Gamma = value;
    tridentSetVideoGamma(pTrident,value,pPriv->Brightness);
  } else if (attribute == xvContrast) {
    /*if ((value < 0) || (value > 127))
      return BadValue;*/
    pPriv->Contrast = value;
    tridentSetVideoContrast(pTrident,value);
  } 
#if TRIDENT_VIP
    else if( attribute == xvEncoding)
    {
        pPriv->encoding = value;
    }
    else if( attribute == xvFrequency)
    {
        pPriv->frequency = value;
    }
    else if( attribute == xvMute)
    {
        pPriv->mute = value;
    }
    else if( attribute == xvVolume)
    {
        pPriv->volume = value;
    }
    else
    {    
        fprintf(stderr,"SetAttribute:%d\n",attribute);    
    }    
#else
  else 
    return BadMatch;
#endif

  return Success;
}

static int 
TRIDENTGetPortAttribute(
  ScrnInfoPtr pScrn, 
  Atom attribute,
  INT32 *value, 
  pointer data
){
  TRIDENTPortPrivPtr pPriv = (TRIDENTPortPrivPtr)data;

#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTGetPortAttribute:\n");
#endif

  if(attribute == xvColorKey) {
	*value = pPriv->colorKey;
  } else if(attribute == xvBrightness) {
	*value = pPriv->Brightness;
  } else if(attribute == xvSaturation) {
	*value = pPriv->Saturation;
  } else if (attribute == xvHue) {
	*value = pPriv->HUE;
  } else if (attribute == xvGamma) {
	*value = pPriv->Gamma;
  } else if (attribute == xvContrast) {
	*value = pPriv->Contrast;
  }

#if TRIDENT_VIP
    else if( attribute == xvDoubleBuffer) 
    {
        *value = pPriv->doubleBuffer ? 1:0;
    }
    else if( attribute == xvEncoding)
    {
        *value = pPriv->encoding;
    }
    else if( attribute == xvFrequency)
    {
        pPriv->frequency=1000;
        *value = pPriv->frequency;
    }
    else if( attribute == xvMute)
    {
        pPriv->mute=1;
        *value = pPriv->mute;
    }
    else if( attribute == xvVolume)
    {
        pPriv->volume=0;
        *value = pPriv->volume;
    }
    else
    {
        #if Video_DEBUG
            fprintf(stderr,"GetAttribute:%d\n",attribute);
        #endif    
    }
    
#else
  else
    return BadMatch;
#endif

  return Success;
}

static void
TRIDENTQueryBestSize(
  ScrnInfoPtr pScrn,
  Bool motion,
  short vid_w, short vid_h,
  short drw_w, short drw_h,
  unsigned int *p_w, unsigned int *p_h,
  pointer data
){
#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTQueryBestSize:\n");
#endif
  *p_w = drw_w;
  *p_h = drw_h;

  if(*p_w > 16384) *p_w = 16384;
}


static void
TRIDENTCopyData(
  unsigned char *src,
  unsigned char *dst,
  int srcPitch,
  int dstPitch,
  int h,
  int w
){
    w <<= 1;
    while(h--) {
	memcpy(dst, src, w);
	src += srcPitch;
	dst += dstPitch;
    }
}

static void
TRIDENTCopyMungedData(
   unsigned char *src1,
   unsigned char *src2,
   unsigned char *src3,
   unsigned char *dst1,
   int srcPitch,
   int srcPitch2,
   int dstPitch,
   int h,
   int w
){
   CARD32 *dst = (CARD32*)dst1;
   int i, j;

   dstPitch >>= 2;
   w >>= 1;

   for(j = 0; j < h; j++) {
	for(i = 0; i < w; i++) {
	    dst[i] = src1[i << 1] | (src1[(i << 1) + 1] << 16) |
		     (src3[i] << 8) | (src2[i] << 24);
	}
	dst += dstPitch;
	src1 += srcPitch;
	if(j & 1) {
	    src2 += srcPitch2;
	    src3 += srcPitch2;
	}
   }
}

static FBLinearPtr
TRIDENTAllocateMemory(
   ScrnInfoPtr pScrn,
   FBLinearPtr linear,
   int size
){
   ScreenPtr pScreen;
   FBLinearPtr new_linear;

   if(linear) {
	if(linear->size >= size)
	   return linear;

        if(xf86ResizeOffscreenLinear(linear, size))
	   return linear;

	xf86FreeOffscreenLinear(linear);
   }

   pScreen = screenInfo.screens[pScrn->scrnIndex];

   new_linear = xf86AllocateOffscreenLinear(pScreen, size, 16,
   						NULL, NULL, NULL);

   if(!new_linear) {
	int max_size;

	xf86QueryLargestOffscreenLinear(pScreen, &max_size, 16, 
						PRIORITY_EXTREME);
	
	if(max_size < size)
	   return NULL;

	xf86PurgeUnlockedOffscreenAreas(pScreen);
	new_linear = xf86AllocateOffscreenLinear(pScreen, size, 16, 
						NULL, NULL, NULL);
   }

   return new_linear;
}

static void
TRIDENTDisplayVideo(
    ScrnInfoPtr pScrn,
    int id,
    int offset,
    short width, short height,
    int pitch,
    int x1, int y1, int x2, int y2,
    BoxPtr dstBox,
    short src_w, short src_h,
    short drw_w, short drw_h
){
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    int vgaIOBase = VGAHWPTR(pScrn)->IOBase;
    int zoomx1, zoomx2, zoomy1, zoomy2;
    int tx1,tx2;
    int ty1,ty2;

#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTDisplayVideo:\n");
#endif

    switch(id) {
    case 0x35315652:		/* RGB15 */
    case 0x36315652:		/* RGB16 */
	if (pTrident->Chipset >= CYBER9397) {
    	    OUTW(vgaIOBase + 4, 0x22BF);
	} else {
    	    OUTW(vgaIOBase + 4, 0x118F);
	}
	break;
    case FOURCC_YV12:		/* YV12 */
    case FOURCC_YUY2:		/* YUY2 */
    default:
	if (pTrident->Chipset >= CYBER9397) {
    	    OUTW(vgaIOBase + 4, 0x00BF);
	} else {
    	    OUTW(vgaIOBase + 4, 0x108F);
	}
	break;
    }
    
    tx1 = dstBox->x1 + pTrident->hsync;    
    tx2 = dstBox->x2 + pTrident->hsync + pTrident->hsync_rskew;
    ty1 = dstBox->y1 + pTrident->vsync - 2;
    ty2 = dstBox->y2 + pTrident->vsync + 2 + pTrident->vsync_bskew;

    OUTW(vgaIOBase + 4, (tx1 & 0xff) <<8 | 0x86);
    OUTW(vgaIOBase + 4, (tx1 & 0xff00)   | 0x87);
    OUTW(vgaIOBase + 4, (ty1 & 0xff) <<8 | 0x88);
    OUTW(vgaIOBase + 4, (ty1 & 0xff00)   | 0x89);
    OUTW(vgaIOBase + 4, (tx2 & 0xff) <<8 | 0x8A);
    OUTW(vgaIOBase + 4, (tx2 & 0xff00)   | 0x8B);
    OUTW(vgaIOBase + 4, (ty2 & 0xff) <<8 | 0x8C);
    OUTW(vgaIOBase + 4, (ty2 & 0xff00)   | 0x8D);
    
    offset += (x1 >> 15) & ~0x01;

    if (pTrident->videoFlags & VID_OFF_SHIFT_4)
    {
        /*ErrorF("VIAB3D: TRIDENTDisplayVideo: VID_OFF_SHIFT_4 -> offset=%04x\n", offset);*/
        offset = offset >> 4;
    }
    else
    {
        /*ErrorF("VIAB3D: TRIDENTDisplayVideo: VID_OFF_SHIFT_3 -> offset=%04x\n", offset);*/
        offset = offset >> 3;
    }

    OUTW(vgaIOBase + 4, (((width<<1) & 0xff)<<8)   | 0x90);
    OUTW(vgaIOBase + 4, ((width<<1) & 0xff00)      | 0x91);
    OUTW(vgaIOBase + 4, ((offset) & 0xff) << 8     | 0x92);
    OUTW(vgaIOBase + 4, ((offset) & 0xff00)        | 0x93);
    OUTW(vgaIOBase + 4, ((offset) & 0xff0000) >> 8 | 0x94);

    /* Horizontal Zoom */
    if (pTrident->videoFlags & VID_ZOOM_INV) 
    {
#if TRIDENT_VIP
       if (bVIP_ON) /* Adjust for VIP zoom horizontal factor */
       {
         /*ErrorF("VIAB3D: TRIDENTDisplayVideo: minify=> src_w=%d, drw_w=%d\n", src_w, drw_w);*/
         drw_w += 13;
       }
#endif  
      if ((pTrident->videoFlags & VID_ZOOM_MINI) && src_w > drw_w)
	    {
	      zoomx2 = (int)((float)drw_w/(float)src_w * 1024) 
	        | (((int)((float)src_w/(float)drw_w) - 1)&7)<<10 | 0x8000;
    	}
      else
	    {
	      zoomx2 = (int)(float)(src_w-1)/(float)(drw_w+3) * 1024; /* Ian */
	    }
	    /*ErrorF("VIAB3D: TRIDENTDisplayVideo: zoomx2=%04x\n", zoomx2);*/
      OUTW(vgaIOBase + 4, (zoomx2&0xff)<<8 | 0x80);
	    OUTW(vgaIOBase + 4, (zoomx2&0x9f00) | 0x81);
  }  
  else
  {
    if (drw_w == src_w) 
    {
  	  OUTW(vgaIOBase + 4, 0x0080);
	    OUTW(vgaIOBase + 4, 0x0081);
    } 
    else if (drw_w > src_w) 
    {
	    float z;
      z = (float)drw_w/(float)src_w - 1;
	    zoomx1 =  z;
	    zoomx2 = (z - (int)zoomx1 ) * 1024;
		
	    OUTW(vgaIOBase + 4, (zoomx2&0xff)<<8 | 0x80);
		  OUTW(vgaIOBase + 4, (zoomx1&0x0f)<<10 | (zoomx2&0x0300) |0x81);
    }
    else
    {
	    zoomx1 =   ((float)drw_w/(float)src_w);
	    zoomx2 = ( ((float)drw_w/(float)src_w) - (int)zoomx1 ) * 1024;
	    OUTW(vgaIOBase + 4, (zoomx2&0xff)<<8 |   0x80);
	    OUTW(vgaIOBase + 4, (zoomx2&0x0300)|
			(((int)((float)src_w/(float)drw_w)-1)&7)<<10 | 0x8081);
    }
  }

    /* Vertical Zoom */
    if (pTrident->videoFlags & VID_ZOOM_INV) {
       if ((pTrident->videoFlags & VID_ZOOM_MINI) && src_h > drw_h)
         zoomy2 = (int)(( ((float)drw_h/(float)src_h)) * 1024) 
	   | (((int)((float)src_h/(float)drw_h)-1)&7)<<10 
	   | 0x8000;
       else
         zoomy2 = ( ((float)(src_h-1)/(float)(drw_h+1))) * 1024; /* Ian */
	OUTW(vgaIOBase + 4, (zoomy2&0xff)<<8 | 0x82);
	OUTW(vgaIOBase + 4, (zoomy2&0x9f00) | 0x0083);
    } else {
    if (drw_h == src_h) {
	OUTW(vgaIOBase + 4, 0x0082);
	OUTW(vgaIOBase + 4, 0x0083);
    } else
    if (drw_h > src_h) {
	float z;

	z = (float)drw_h/(float)src_h - 1;
	zoomy1 =  z;
	zoomy2 = (z - (int)zoomy1 ) * 1024;

	OUTW(vgaIOBase + 4, (zoomy2&0xff)<<8 | 0x82);
		OUTW(vgaIOBase + 4, (zoomy1&0x0f)<<10 | (zoomy2&0x0300) |0x83);
    } else {
	zoomy1 =   ((float)drw_h/(float)src_h);
	zoomy2 = ( ((float)drw_h/(float)src_h) - (int)zoomy1 ) * 1024;
	OUTW(vgaIOBase + 4, (zoomy2&0xff)<<8 | 0x82);
	OUTW(vgaIOBase + 4, (zoomy2&0x0300)|
			(((int)((float)src_h/(float)drw_h)-1)&7)<<10 | 0x8083);
    }
    } 

    if (pTrident->Chipset >= CYBER9397) {
	int lb = (width+2) >> 2;

    	OUTW(vgaIOBase + 4, ((lb & 0x100)>>1) | 0x0895);
    	OUTW(vgaIOBase + 4,  (lb & 0xFF)<<8   | 0x0096);

    	if (src_w > 384) { 
    	    OUTW(0x3C4, 0x0497); /* 2x line buffers */ 
    	} else {
    	    OUTW(0x3C4, 0x0097); /* 1x line buffers */
    	}
    	OUTW(vgaIOBase + 4, 0x0097); 
    	OUTW(vgaIOBase + 4, 0x00BA);
    	OUTW(vgaIOBase + 4, 0x00BB);
    	OUTW(vgaIOBase + 4, 0xFFBC);
    	OUTW(vgaIOBase + 4, 0xFFBD);
    	OUTW(vgaIOBase + 4, 0x04BE); 
    	OUTW(vgaIOBase + 4, 0xD48E);
    	OUTW(vgaIOBase + 4, 0x208F);
    } else {
	
    	OUTW(vgaIOBase + 4, ((((id == FOURCC_YV12) || (id == FOURCC_YUY2)) 
					? (width >> 2) : (width >> 6)) << 8) | 0x95);
    	OUTW(vgaIOBase + 4, ((((id == FOURCC_YV12) || (id == FOURCC_YUY2)) 
				? ((width+2) >> 2) : ((width+2) >> 6)) << 8) |0x96);

    	OUTW(vgaIOBase + 4, 0x938E);
	OUTB(0x83C8, 0x00);
	OUTB(0x83C6, 0x95);
    }
}

static int 
TRIDENTPutImage( 
  ScrnInfoPtr pScrn, 
  short src_x, short src_y, 
  short drw_x, short drw_y,
  short src_w, short src_h, 
  short drw_w, short drw_h,
  int id, unsigned char* buf, 
  short width, short height, 
  Bool sync,
  RegionPtr clipBoxes, pointer data
){
   TRIDENTPortPrivPtr pPriv = (TRIDENTPortPrivPtr)data;
   TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
   INT32 x1, x2, y1, y2;
   unsigned char *dst_start;
   int pitch, new_size, offset, offset2 = 0, offset3 = 0;
   int srcPitch, srcPitch2 = 0, dstPitch;
   int top, left, npixels, nlines, bpp;
   BoxRec dstBox;
   CARD32 tmp;
   
#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTPutImage:\n");
#endif

   /* Clip */
   x1 = src_x;
   x2 = src_x + src_w;
   y1 = src_y;
   y2 = src_y + src_h;

   dstBox.x1 = drw_x;
   dstBox.x2 = drw_x + drw_w;
   dstBox.y1 = drw_y;
   dstBox.y2 = drw_y + drw_h;
   
   if(!xf86XVClipVideoHelper(&dstBox, &x1, &x2, &y1, &y2, clipBoxes, 
								width, height))
	    return Success;   
      
   dstBox.x1 -= pScrn->frameX0;   
   dstBox.x2 -= pScrn->frameX0;
   dstBox.y1 -= pScrn->frameY0;
   dstBox.y2 -= pScrn->frameY0;

   bpp = pScrn->bitsPerPixel >> 3;
   pitch = bpp * pScrn->displayWidth;

   dstPitch = ((width << 1) + 15) & ~15;
   new_size = ((dstPitch * height) + bpp - 1) / bpp;
   switch(id) {
   case FOURCC_YV12:
   case FOURCC_I420:
	srcPitch = (width + 3) & ~3;
	offset2 = srcPitch * height;
	srcPitch2 = ((width >> 1) + 3) & ~3;
	offset3 = (srcPitch2 * (height >> 1)) + offset2;
	break;
   case FOURCC_UYVY:
   case FOURCC_YUY2:
   default:
	srcPitch = (width << 1);
	break;
   }  

   if(!(pPriv->linear = TRIDENTAllocateMemory(pScrn, pPriv->linear, new_size)))
	return BadAlloc;

    /* copy data */
   top = y1 >> 16;
   left = (x1 >> 16) & ~1;
   npixels = ((((x2 + 0xffff) >> 16) + 1) & ~1) - left;
   left <<= 1;

   offset = pPriv->linear->offset * bpp;

   dst_start = pTrident->FbBase + offset + left + (top * dstPitch);

   switch(id) {
    case FOURCC_YV12:
    case FOURCC_I420:
	top &= ~1;
	tmp = ((top >> 1) * srcPitch2) + (left >> 2);
	offset2 += tmp;
	offset3 += tmp;
	if(id == FOURCC_I420) {
	   tmp = offset2;
	   offset2 = offset3;
	   offset3 = tmp;
	}
	nlines = ((((y2 + 0xffff) >> 16) + 1) & ~1) - top;
	TRIDENTCopyMungedData(buf + (top * srcPitch) + (left >> 1), 
			  buf + offset2, buf + offset3, dst_start,
			  srcPitch, srcPitch2, dstPitch, nlines, npixels);
	break;
    case FOURCC_UYVY:
    case FOURCC_YUY2:
    default:
	buf += (top * srcPitch) + left;
	nlines = ((y2 + 0xffff) >> 16) - top;
	TRIDENTCopyData(buf, dst_start, srcPitch, dstPitch, nlines, npixels);
        break;
    }

    /* update cliplist */
    if(!RegionsEqual(&pPriv->clip, clipBoxes)) {
    	/* update cliplist */
        REGION_COPY(pScreen, &pPriv->clip, clipBoxes);
        xf86XVFillKeyHelper(pScrn->pScreen, pPriv->colorKey, clipBoxes);
    }

    offset += top * dstPitch;

    tridentFixFrame(pScrn,&pPriv->fixFrame);
    TRIDENTDisplayVideo(pScrn, id, offset, width, height, dstPitch,
	     x1, y1, x2, y2, &dstBox, src_w, src_h, drw_w, drw_h);

    pPriv->videoStatus = CLIENT_VIDEO_ON;

    return Success;
}

#if TRIDENT_VIP

static int 
TRIDENTPutVideo( 
    ScrnInfoPtr pScrn, 
    short src_x, short src_y, 
    short drw_x, short drw_y,
    short src_w, short src_h, 
    short drw_w, short drw_h,
    RegionPtr clipBoxes, pointer data
)
{
   TRIDENTPortPrivPtr pPriv = (TRIDENTPortPrivPtr)data;
   TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
   INT32 x1, x2, y1, y2;
   unsigned char *dst_start;
   int pitch, new_size = 0, offset, offset2 = 0, offset3 = 0;
   int srcPitch = 0, srcPitch2 = 0, dstPitch = 0;
   int top, left, npixels, nlines, bpp;
   BoxRec dstBox;
   CARD32 tmp;
   int width,height,id;
   
#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTPutVideo\n");
#endif

   /* Clip */
   x1 = src_x;
   x2 = src_x + src_w;
   y1 = src_y;
   y2 = src_y + src_h;

   dstBox.x1 = drw_x;
   dstBox.x2 = drw_x + drw_w;
   dstBox.y1 = drw_y;
   dstBox.y2 = drw_y + drw_h;

   width = InputVideoEncodings[pPriv->encoding].width;
   height = InputVideoEncodings[pPriv->encoding].height;      
   xf86XVClipVideoHelper(&dstBox, &x1, &x2, &y1, &y2, 
	                        clipBoxes, width, height);		 
   
   dstBox.x1 -= pScrn->frameX0;   
   dstBox.x2 -= pScrn->frameX0;
   dstBox.y1 -= pScrn->frameY0;
   dstBox.y2 -= pScrn->frameY0;

   bpp = pScrn->bitsPerPixel >> 3;
   pitch = bpp * pScrn->displayWidth;

   dstPitch = ((width << 1) + 15) & ~15;
   new_size = ((dstPitch * height) + bpp - 1) / bpp;
   
   id = FOURCC_YUY2;
   switch(id) {
   case FOURCC_YV12:
   case FOURCC_I420:
	srcPitch = (width + 3) & ~3;
	offset2 = srcPitch * height;
	srcPitch2 = ((width >> 1) + 3) & ~3;
	offset3 = (srcPitch2 * (height >> 1)) + offset2;
	break;
   case FOURCC_UYVY:
   case FOURCC_YUY2:
   default:
	srcPitch = (width << 1);
	break;
   }  

   if(!(pPriv->linear = TRIDENTAllocateMemory(pScrn, pPriv->linear, new_size)))
	return BadAlloc;

    /* copy data */
   top = y1 >> 16;
   left = (x1 >> 16) & ~1;
   npixels = ((((x2 + 0xffff) >> 16) + 1) & ~1) - left;
   left <<= 1;

   offset = pPriv->linear->offset * bpp;
   
   dst_start = pTrident->FbBase + offset + left + (top * dstPitch);
    
   switch(id) {
    case FOURCC_YV12:
    case FOURCC_I420:
	top &= ~1;
	tmp = ((top >> 1) * srcPitch2) + (left >> 2);
	offset2 += tmp;
	offset3 += tmp;
	if(id == FOURCC_I420) {
	   tmp = offset2;
	   offset2 = offset3;
	   offset3 = tmp;
	}
	nlines = ((((y2 + 0xffff) >> 16) + 1) & ~1) - top;
	/*TRIDENTCopyMungedData(buf + (top * srcPitch) + (left >> 1), 
			  buf + offset2, buf + offset3, dst_start,
			  srcPitch, srcPitch2, dstPitch, nlines, npixels);*/
	break;
    case FOURCC_UYVY:
    case FOURCC_YUY2:
    default:
	/*buf += (top * srcPitch) + left;
	nlines = ((y2 + 0xffff) >> 16) - top;
	TRIDENTCopyData(buf, dst_start, srcPitch, dstPitch, nlines, npixels);*/
        break;
    }

    /* update cliplist */
    if(!RegionsEqual(&pPriv->clip, clipBoxes)) {
    	/* update cliplist */
        REGION_COPY(pScreen, &pPriv->clip, clipBoxes);
        xf86XVFillKeyHelper(pScrn->pScreen, pPriv->colorKey, clipBoxes);
    }

    offset += top * dstPitch;

    /* Louis: Enable the VIP port */    
    if (!bVIP_Init)  /*Init the VIP for just once */
    {       
      ResetCapturePort();
      InitializeCapturePort();
      bVIP_Init = 1;
    }
    SetMinifyFactor(src_w, src_h, drw_w, drw_h);
    CapturePortSetting(pScrn, offset);
    EnableCapturePort();
        
    tridentFixFrame(pScrn,&pPriv->fixFrame);
    TRIDENTDisplayVideo(pScrn, id, offset, width, height, dstPitch,
	     x1, y1, x2, y2, &dstBox, src_w, src_h, drw_w, drw_h);

    pPriv->videoStatus = CLIENT_VIDEO_ON;

    return Success;
}
#endif

static int
TRIDENTQueryImageAttributes(
  ScrnInfoPtr pScrn,
  int id,
  unsigned short *w, unsigned short *h,
  int *pitches, int *offsets
){
    int size, tmp;

#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTQueryImageAttributes:\n");
#endif

    if(*w > 1024) *w = 1024;
    if(*h > 1024) *h = 1024;

    *w = (*w + 1) & ~1;
    if(offsets) offsets[0] = 0;

    switch(id) {
    case FOURCC_YV12:		/* YV12 */
	*h = (*h + 1) & ~1;
	size = (*w + 3) & ~3;
	if(pitches) pitches[0] = size;
	size *= *h;
	if(offsets) offsets[1] = size;
	tmp = ((*w >> 1) + 3) & ~3;
	if(pitches) pitches[1] = pitches[2] = tmp;
	tmp *= (*h >> 1);
	size += tmp;
	if(offsets) offsets[2] = size;
	size += tmp;
	break;
    default:			/* RGB15, RGB16, YUY2 */
	size = *w << 1;
	if(pitches) pitches[0] = size;
	size *= *h;
	break;
    }

    return size;
}

/****************** Offscreen stuff ***************/

typedef struct {
  FBLinearPtr linear;
  Bool isOn;
} OffscreenPrivRec, * OffscreenPrivPtr;

static int 
TRIDENTAllocateSurface(
    ScrnInfoPtr pScrn,
    int id,
    unsigned short w, 	
    unsigned short h,
    XF86SurfacePtr surface
){
    FBLinearPtr linear;
    int pitch, fbpitch, size, bpp;
    OffscreenPrivPtr pPriv;

    if((w > 1024) || (h > 1024))
	return BadAlloc;

    w = (w + 1) & ~1;
    pitch = ((w << 1) + 15) & ~15;
    bpp = pScrn->bitsPerPixel >> 3;
    fbpitch = bpp * pScrn->displayWidth;
    size = ((pitch * h) + bpp - 1) / bpp;

    if(!(linear = TRIDENTAllocateMemory(pScrn, NULL, size)))
	return BadAlloc;

    surface->width = w;
    surface->height = h;

    if(!(surface->pitches = xalloc(sizeof(int)))) {
	xf86FreeOffscreenLinear(linear);
	return BadAlloc;
    }
    if(!(surface->offsets = xalloc(sizeof(int)))) {
	xfree(surface->pitches);
	xf86FreeOffscreenLinear(linear);
	return BadAlloc;
    }
    if(!(pPriv = xalloc(sizeof(OffscreenPrivRec)))) {
	xfree(surface->pitches);
	xfree(surface->offsets);
	xf86FreeOffscreenLinear(linear);
	return BadAlloc;
    }

    pPriv->linear = linear;
    pPriv->isOn = FALSE;

    surface->pScrn = pScrn;
    surface->id = id;   
    surface->pitches[0] = pitch;
    surface->offsets[0] = linear->offset * bpp;
    surface->devPrivate.ptr = (pointer)pPriv;

    return Success;
}

static int
TRIDENTStopSurface(
    XF86SurfacePtr surface
){
    OffscreenPrivPtr pPriv = (OffscreenPrivPtr)surface->devPrivate.ptr;

    if(pPriv->isOn) {
	TRIDENTPtr pTrident = TRIDENTPTR(surface->pScrn);
    	int vgaIOBase = VGAHWPTR(surface->pScrn)->IOBase;
	OUTW(vgaIOBase + 4, 0x008E);
	OUTW(vgaIOBase + 4, 0x008F);
	pPriv->isOn = FALSE;
    }

    return Success;
}


static int 
TRIDENTFreeSurface(
    XF86SurfacePtr surface
){
    OffscreenPrivPtr pPriv = (OffscreenPrivPtr)surface->devPrivate.ptr;

    if(pPriv->isOn)
	TRIDENTStopSurface(surface);
    xf86FreeOffscreenLinear(pPriv->linear);
    xfree(surface->pitches);
    xfree(surface->offsets);
    xfree(surface->devPrivate.ptr);

    return Success;
}

static int
TRIDENTGetSurfaceAttribute(
    ScrnInfoPtr pScrn,
    Atom attribute,
    INT32 *value
){
    return TRIDENTGetPortAttribute(pScrn, attribute, value, 
			(pointer)(GET_PORT_PRIVATE(pScrn)));
}

static int
TRIDENTSetSurfaceAttribute(
    ScrnInfoPtr pScrn,
    Atom attribute,
    INT32 value
){
    return TRIDENTSetPortAttribute(pScrn, attribute, value, 
			(pointer)(GET_PORT_PRIVATE(pScrn)));
}

static int 
TRIDENTDisplaySurface(
    XF86SurfacePtr surface,
    short src_x, short src_y, 
    short drw_x, short drw_y,
    short src_w, short src_h, 
    short drw_w, short drw_h,
    RegionPtr clipBoxes
){
    OffscreenPrivPtr pPriv = (OffscreenPrivPtr)surface->devPrivate.ptr;
    ScrnInfoPtr pScrn = surface->pScrn;
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    TRIDENTPortPrivPtr portPriv = pTrident->adaptor->pPortPrivates[0].ptr;
    INT32 x1, y1, x2, y2;
    BoxRec dstBox;

#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTDisplaySurface:\n");
#endif

    x1 = src_x;
    x2 = src_x + src_w;
    y1 = src_y;
    y2 = src_y + src_h;

    dstBox.x1 = drw_x;
    dstBox.x2 = drw_x + drw_w;
    dstBox.y1 = drw_y;
    dstBox.y2 = drw_y + drw_h;

    if(!xf86XVClipVideoHelper(&dstBox, &x1, &x2, &y1, &y2, clipBoxes,
			surface->width, surface->height))
    {
	return Success;
    }

    dstBox.x1 -= pScrn->frameX0;
    dstBox.x2 -= pScrn->frameX0;
    dstBox.y1 -= pScrn->frameY0;
    dstBox.y2 -= pScrn->frameY0;

    TRIDENTResetVideo(pScrn);

    tridentFixFrame(pScrn,&portPriv->fixFrame);
    TRIDENTDisplayVideo(pScrn, surface->id, surface->offsets[0], 
	     surface->width, surface->height, surface->pitches[0],
	     x1, y1, x2, y2, &dstBox, src_w, src_h, drw_w, drw_h);

    xf86XVFillKeyHelper(pScrn->pScreen, portPriv->colorKey, clipBoxes);

    pPriv->isOn = TRUE;
    /* we've prempted the XvImage stream so set its free timer */
    if(portPriv->videoStatus & CLIENT_VIDEO_ON) {
	REGION_EMPTY(pScrn->pScreen, &portPriv->clip);   
	UpdateCurrentTime();
	portPriv->videoStatus = FREE_TIMER;
	portPriv->freeTime = currentTime.milliseconds + FREE_DELAY;
	pTrident->VideoTimerCallback = TRIDENTVideoTimerCallback;
    }

    return Success;
}

static void
TRIDENTInitOffscreenImages(ScreenPtr pScreen)
{
    XF86OffscreenImagePtr offscreenImages;

#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTInitOffscreenImages:\n");
#endif

    /* need to free this someplace */
    if(!(offscreenImages = xalloc(sizeof(XF86OffscreenImageRec))))
	return;

    offscreenImages[0].image = &Images[0];
    offscreenImages[0].flags = VIDEO_OVERLAID_IMAGES | 
			       VIDEO_CLIP_TO_VIEWPORT;
    offscreenImages[0].alloc_surface = TRIDENTAllocateSurface;
    offscreenImages[0].free_surface = TRIDENTFreeSurface;
    offscreenImages[0].display = TRIDENTDisplaySurface;
    offscreenImages[0].stop = TRIDENTStopSurface;
    offscreenImages[0].setAttribute = TRIDENTSetSurfaceAttribute;
    offscreenImages[0].getAttribute = TRIDENTGetSurfaceAttribute;
    offscreenImages[0].max_width = 1024;
    offscreenImages[0].max_height = 1024;
    offscreenImages[0].num_attributes = NUM_ATTRIBUTES;
    offscreenImages[0].attributes = Attributes;
    
    xf86XVRegisterOffscreenImages(pScreen, offscreenImages, 1);
}

static void
TRIDENTVideoTimerCallback(ScrnInfoPtr pScrn, Time time)
{
    TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    TRIDENTPortPrivPtr pPriv = pTrident->adaptor->pPortPrivates[0].ptr;
    int vgaIOBase = VGAHWPTR(pScrn)->IOBase;

#if Video_DEBUG
    ErrorF("VIAB3D: TRIDENTVideoTimerCallback:\n");
#endif

    if(pPriv->videoStatus & TIMER_MASK) {
	if(pPriv->videoStatus & OFF_TIMER) {
	    if(pPriv->offTime < time) {
		OUTW(vgaIOBase + 4, 0x008E);
		OUTW(vgaIOBase + 4, 0x008F);
		pPriv->videoStatus = FREE_TIMER;
		pPriv->freeTime = time + FREE_DELAY;
	    }
	} else {  /* FREE_TIMER */
	    if(pPriv->freeTime < time) {
		if(pPriv->linear) {
		   xf86FreeOffscreenLinear(pPriv->linear);
		   pPriv->linear = NULL;
		}
		pPriv->videoStatus = 0;
	        pTrident->VideoTimerCallback = NULL;
	    }
        }
    } else  /* shouldn't get here */
	pTrident->VideoTimerCallback = NULL;
}

    /* Calculate skew offsets for video overlay */


void
tridentFixFrame(ScrnInfoPtr pScrn, int *fixFrame)
{

  TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
  int vgaIOBase = VGAHWPTR(pScrn)->IOBase;
  int HTotal, HSyncStart;
  int VTotal, VSyncStart;
  int h_off = 0;
  int v_off = 0;
  unsigned char CRTC[0x11];
  Bool isShadow;
  unsigned char shadow = 0;

#if Video_DEBUG
    ErrorF("VIAB3D: tridentFixFrame:\n");
#endif

  if ((*fixFrame)++ < 100)
    return;

  *fixFrame = 0;
  isShadow = (pTrident->ModeReg.tridentRegs3CE[CyberControl] & 0x81) == 0x81;

  if (isShadow)
    SHADOW_ENABLE(shadow);

  OUTB(vgaIOBase + 4, 0x0);
  CRTC[0x0] = INB(vgaIOBase + 5);
  OUTB(vgaIOBase + 4, 0x4);
  CRTC[0x4] = INB(vgaIOBase + 5);
  OUTB(vgaIOBase + 4, 0x5);
  CRTC[0x5] = INB(vgaIOBase + 5);
  OUTB(vgaIOBase + 4, 0x6);
  CRTC[0x6] = INB(vgaIOBase + 5);
  OUTB(vgaIOBase + 4, 0x7);
  CRTC[0x7] = INB(vgaIOBase + 5);
  OUTB(vgaIOBase + 4, 0x10);
  CRTC[0x10] = INB(vgaIOBase + 5);

  HTotal = CRTC[0] << 3;
  VTotal = CRTC[6] 
    | ((CRTC[7] & (1<<0)) << 8)
    | ((CRTC[7] & (1<<5)) << 4);
  HSyncStart = (CRTC[4] 
		+ ((CRTC[5] >> 5) & 0x3)) << 3;
  VSyncStart = CRTC[0x10] 
    | ((CRTC[7] & (1<<2)) << 6)
    | ((CRTC[7] & (1<<7)) << 2);

  if (isShadow) {
    SHADOW_RESTORE(shadow);
    if (pTrident->lcdMode != 0xff) {
      h_off = (LCD[pTrident->lcdMode].display_x 
	       - pScrn->currentMode->HDisplay) >> 1;
      v_off = (LCD[pTrident->lcdMode].display_y 
	       - pScrn->currentMode->VDisplay) >> 1;
    }
  } 

  pTrident->hsync = (HTotal - HSyncStart) + 23 + h_off;
  pTrident->vsync = (VTotal - VSyncStart) - 2 + v_off;
  
  /* 
   * HACK !! As awful as this is, it appears to be the only way....Sigh!
   * We have XvHsync and XvVsync as options now, which adjust 
   * at the very end of this function. It'll be helpful for now
   * and we can get more data on some of these skew values.
   */
  switch (pTrident->Chipset) {
  case TGUI9680:
    /* Furthur tweaking needed */
#if 0
    pTrident->hsync -= (mode->CrtcHTotal / 16);
#endif
    pTrident->vsync += 2;
    break;
  case PROVIDIA9682:
    /* Furthur tweaking needed */
    pTrident->hsync += 7;
    break;
  case PROVIDIA9685:
    /* Spot on */
    break;
  case CYBERBLADEXPm8:
  case CYBERBLADEXPm16:
  case CYBERBLADEXPAI1:
    pTrident->hsync -= 15;
    pTrident->hsync_rskew = 3;
    break;
  case BLADE3D:
    if (pScrn->depth == 24)
      pTrident->hsync -= 8;
    else
      pTrident->hsync -= 6;
    break;
  case CYBERBLADEI7:
  case CYBERBLADEI7D:
  case CYBERBLADEI1:
  case CYBERBLADEI1D:
    pTrident->hsync -= 8;
    break;
  case CYBERBLADEAI1:
    pTrident->hsync -= 7;
    break;
  case CYBERBLADEAI1D:
      pTrident->vsync += 2;
      pTrident->vsync_bskew = -4;
      pTrident->hsync -= 5;
      break;
  case CYBERBLADEE4:
    pTrident->hsync -= 8;
    break;
  }
  pTrident->hsync+=pTrident->OverrideHsync;
  pTrident->vsync+=pTrident->OverrideVsync;
  pTrident->hsync_rskew+=pTrident->OverrideRskew;
  pTrident->vsync_bskew+=pTrident->OverrideBskew;
}
    
#if TRIDENT_VIP

unsigned char getCR(unsigned char  nIndex)
{    
    outb(0x3d4, nIndex);    
    return inb(0x3d5);
}

unsigned char getSR(unsigned char nIndex)
{
    outb(0x3c4, nIndex);    
    return inb(0x3c5);
}

unsigned char getGR(unsigned char nIndex)
{
    outb(0x3ce, nIndex);    
    return inb(0x3cf);
}

void setCR(unsigned char nIndex, unsigned char nValue)
{
    outb(0x3d4, nIndex);
    outb(0x3d5, nValue);    
}

void setSR(unsigned char nIndex, unsigned char nValue)
{
    outb(0x3c4, nIndex);
    outb(0x3c5, nValue);
}

void setGR(unsigned char nIndex, unsigned char nValue)
{
    outb(0x3ce, nIndex);
    outb(0x3cf, nValue);
}

/*
 *  Set Video Port to default value
 */
void ResetCapturePort(void){
    unsigned char ucTempValue;

#if Video_DEBUG
    ErrorF("VIAB3D: ResetCapturePort\n");
#endif

	  /* Unprotect all extended register by setting 3c5.11 to 0x92*/
  	outw(0x9211,0x3c4);
    
	  /* Motion Video Port Suspend - Disable GR2a*/	  
	  ucTempValue =  getGR(0x2a) & ~0x80;
	  setGR(0x2a, ucTempValue);
    
	  ucTempValue =  getCR(0x9c);    
    ucTempValue |= 0x10; 
    ucTempValue |= 0x40; 
    setCR(0x9c, ucTempValue);	  
	  setCR(0x9d, 0x00);
	  setCR(0x9e, 0x00);
  	setCR(0x9f, 0x00);
  	setCR(0xae, 0x83); 	
  	setCR(0xaf, 0x00);
  	setCR(0xb8, 0x51);  	
  	setCR(0xb9, 0x00);
  	  	  	    
	  /* Clean PCI Video Minifier & Video Aperture*/	  
	  ucTempValue =  getCR(0x36);
    ucTempValue &= ~0x0060;
    setCR(0x36, ucTempValue);
    
	  /* Reprotect all extended register by setting 3c5.11 to 87*/
	  outw(0x8711,0x3c4);
}

void InitializeCapturePort(void)
{
#if Video_DEBUG
    ErrorF("VIAB3D: InitializeCapturePort\n");
#endif

    setSR(0x11, 0x92);

    setGR(0x2a, 0x01);

    setSR(0xac, 0xCC);
    setSR(0xad, 0x00);
    setSR(0xae, 0xa0);
    setSR(0xaf, 0x05);
    setSR(0xb0, 0x08);
    setSR(0xb1, 0x00);
    setSR(0xb5, 0xa6);
    setSR(0xb6, 0x0e);
    setSR(0xb7, 0x82);
    setSR(0xb9, 0xa6);
    setSR(0xba, 0x0e);
    setSR(0xbc, 0x07);
    setSR(0xbe, 0x00);
    setCR(0x80, 0x3d);
    setCR(0x81, 0x86);
    setCR(0x82, 0xEE);
    setCR(0x83, 0x83);
    setCR(0x8e, 0xD0);
    setCR(0x8f, 0x00);
    setCR(0x95, 0x08);
    setCR(0x96, 0xB4);
    setCR(0x97, 0x03);
    setCR(0x9c, 0xc3);
    setCR(0x9d, 0x05);
    setCR(0x9e, 0x08);
    setCR(0x9f, 0x40);
    setCR(0xa1, 0x02);
    setCR(0xa3, 0x01);
    setCR(0xa5, 0x00);
    setCR(0xa7, 0x01);
    setCR(0xa9, 0x00);
    setCR(0xab, 0x03);
    setCR(0xae, 0x83);
    setCR(0xaf, 0x90);

    setCR(0xb4, 0xCF);
    setCR(0xb5, 0x02);
    setCR(0xb6, 0xDF);
    setCR(0xb7, 0x01);
    setCR(0xb8, 0x05);
    setCR(0xb9, 0x80);
    setCR(0xbe, 0x26);
    setCR(0xbf, 0x00);
}


/* Louis: Not work, use video minify only 
          For xawtv may strang*/
void SetMinifyFactor(short src_w, short src_h, short drw_w, short drw_h)
{

#if Video_DEBUG
    ErrorF("VIAB3D: SetMinifyFactor\n");
    ErrorF("VIAB3D: SetMinifyFactor : src_w=%d, src_h=%d, drw_w=%d, drw_h=%d\n", src_w, src_h, drw_w, drw_h);
#endif
  
  setCR(0xb0, 0xff); /* CapHorMinifyReg */
	setCR(0xb1, 0x03); 	
	setCR(0xb2, 0xff); /* CapVertMinifyReg */
	setCR(0xb3, 0x03); 	
}

void CapturePortSetting(ScrnInfoPtr pScrn, int offset)
{
	unsigned char    ucTemp;
  TRIDENTPtr pTrident = TRIDENTPTR(pScrn);
    
    
#if Video_DEBUG
    ErrorF("VIAB3D: CapturePortSetting\n");
#endif
  
	/* Unprotect all extended register by setting 3c5.11 to 0x92 */
	outw(0x9211,0x3c4);

  /* Video Engine Flag Setting */
  ucTemp = getCR(0x94) | 0x10;
	setCR(0x94, ucTemp); /* VidEngFlagExReg */

	ucTemp = getCR(0x97) | 0x10 ;
	setCR(0x97, 0x10); /* VidEngFlagReg */

	/* Capture Port Address Setting , Double Buffer Setting */
	/* Buffer1 Setting */	
  if (pTrident->videoFlags & VID_OFF_SHIFT_4)
      offset = offset >> 4;
  else
      offset = offset >> 3;

  ucTemp = offset & 0xff;  
	setCR(0x98, ucTemp);		
	ucTemp = (offset & 0xff00) >> 8;	
  setCR(0x99, ucTemp);    
  ucTemp = (offset & 0x070000) >> 16;
  setCR(0x9a, ucTemp);    
   
  /* Capture Port Ctrl Control Setting */
  /*PortControl : Odd 30fs, flicker free*/
  ucTemp = getCR(0xbe) | 0x2;
	setCR(0xbe, ucTemp);
	
	/*PortControl1: Odd 30fs, CapCtrl_VSYNCLOW, CapCtrl_HSYNCLOW */
	ucTemp = getCR(0x9c) | 0xc0 | 0x01 | 0x02;  
	setCR(0x9c, ucTemp);
	
	/*PortControl2: CapCtrl_8BITDATA, CapMode_YUVSWAP, CapMode_UVSWAP */
	ucTemp = getCR(0x9d) | 0x01 | 0x04 | 0x08;  
	setCR(0x9d, ucTemp);
	
	/*PortControl3: CapMode_YUV422 */
	ucTemp = getCR(0x9e) & (~0xc0);  
	setCR(0x9e, ucTemp);
	
	/*PortControl4: */
	ucTemp = getCR(0x9f) | 0x01;  
	setCR(0x9f, ucTemp);
	
	/*PortControl5: CapCtrl_DOUBLEPCLK */
	ucTemp = getCR(0xaf) | 0x10;  
	setCR(0xaf, ucTemp);

/*Louis: About the value of timing, i use "try and error" 
         It should be different with diffent video decoder
         or different EM84xx driver */

  /* Capture Vertical Timing Setting */
  setCR(0xa0, 0x0d); /* CaptureVerticalTotal */
  setCR(0xa1, 0x02); /* CaptureVerticalTotal */
	setCR(0xa2, 0xb2); /* CaptureHorTotal */
	setCR(0xa3, 0x01); /* CaptureHorTotal */
	setCR(0xa4, 0x21); /* CaptureVertStart */
	setCR(0xa5, 0x0); /* CaptureVertStart */
  /*ucTemp = 518 - 50;
	setCR(0xa6, ucTemp);*/ /* CaptureVertEnd */
	setCR(0xa6, 0x03); /* CaptureVertEnd */
	setCR(0xa7, 0x02); /* CaptureVertEnd */

	/* Capture Horizotal Timing Setting */
	setCR(0xa8, 0x83); /* CapHorStartReg */
	setCR(0xa9, 0x00); /* CapHorStartReg */
	setCR(0xaa, 0xa6); /* CapHorEndReg */
	setCR(0xab, 0x03); /* CapHorEndReg */
	setCR(0xac, 0x06); /* CapVertSyncWidthReg */
	setCR(0xad, 0x20); /* CapVertSyncWidthReg */
	setCR(0xae, 0x83); /* CapHorSyncWidthReg */
	setCR(0xaf, 0x10); /* CapHorSyncWidthReg */

	/* CRTC Control Setting */
	/* CapCtrl_DOUBLEPCLK, CapCtrl_ODDEVENSWAP */
	ucTemp = getCR(0xae) | 0x83 | 0x10;
	setCR(0xae, ucTemp); /* CRTCCtrlReg  */

	setCR(0xb8, 0x51); /* CapFIFOCtrlReg1 */
	setCR(0xb9,	0x00); /* CapFIFOCtrlReg2 */

	/* Video Windows Ctrl Reg */
	ucTemp = getSR(0x97) | 0x04;
	setSR(0x97, ucTemp); /* WindowCtrlReg1 */
	ucTemp = getSR(0x98);
	setSR(0x98, ucTemp); /* WindowCtrlReg2 */

  /* VBI Ctrl Reg */
  /* bug patches disable VBI */
	setSR(0x9f, 0x00); /* VBICtrlReg */

	/* CSC and Pitch Setting */
	/* bug patches disable VBI */
	setSR(0xad, 0x00); /* CapCSC */
            
	ucTemp = getCR(0x9b) | 0xc0;
	setCR(0x9b, ucTemp);
    
	ucTemp = getCR(0xbe) | 0x04;        
	setCR(0xbe, ucTemp);
     
	/* Reprotect all extended register by setting 3c5.11 to 87 */
	outw(0x8711,0x3c4);
}

void EnableCapturePort(void)
{
	unsigned char ucTemp;
#if Video_DEBUG
    ErrorF("VIAB3D: EnableCapturePort\n");
#endif

  bVIP_ON = 1;
  
	/* Unprotect all extended register by setting 3c5.11 to 0x92 */
	outw(0x9211,0x3c4);
	/* Always capture */
	ucTemp =  getCR(0xaf);
  ucTemp |= 0x80; 
	setCR(0xaf, ucTemp);

	ucTemp =  getCR(0x9c);
  ucTemp |= 0x10; 
	setCR(0x9c, ucTemp);
            
	/* Reprotect all extended register by setting 3c5.11 to 87 */
	outw(0x8711,0x3c4);
}

void DisableCapturePort(void)
{
	unsigned char ucTemp;
#if Video_DEBUG
    ErrorF("VIAB3D: DisableCapturePort\n");
#endif

	/* Unprotect all extended register by setting 3c5.11 to 0x92 */
	outw(0x9211,0x3c4);
	/* Disable video capture */
	ucTemp =  getCR(0xaf);
  ucTemp &= 0x7f; 
	setCR(0xaf, ucTemp);

	ucTemp =  getCR(0x9c);
  ucTemp &= 0xef; 
	setCR(0x9c, ucTemp);
            
	/* Reprotect all extended register by setting 3c5.11 to 87 */
	outw(0x8711,0x3c4);
}

#endif

